"""The sphinx extension for the package."""
from __future__ import annotations

from contextlib import suppress
import hashlib
from pathlib import Path, PurePosixPath
import shutil
import subprocess
from time import sleep
import typing as t

from sphinx.application import Sphinx
from sphinx.util import logging

from autodoc2 import __version__
from autodoc2.analysis import analyse_module
from autodoc2.config import CONFIG_PREFIX, Config, ValidationError
from autodoc2.db import InMemoryDb, UniqueError
from autodoc2.utils import ResolvedDict, WarningSubtypes, resolve_all, yield_modules

try:
    from sphinx.util.display import status_iterator
except ImportError:
    # sphinx <6 compatibility
    from sphinx.util import status_iterator  # type: ignore

LOGGER = logging.getLogger("autodoc2")


def warn(msg: str, subtype: WarningSubtypes) -> None:
    """Log a warning."""
    LOGGER.warning(
        f"{msg} [autodoc2.{subtype.value}]", type="autodoc2", subtype=subtype.value
    )


def setup(app: Sphinx) -> dict[str, str | bool]:
    """Entry point for sphinx."""

    # create the configuration options
    for name, default, field in Config().as_triple():
        app.add_config_value(
            f"{CONFIG_PREFIX}{name}",
            field.metadata.get("sphinx_default", default),
            "env",
            types=field.metadata.get("sphinx_type", t.Any),
        )

    # create the main event
    app.connect("builder-inited", run_autodoc)

    # TODO support viewcode, when a package is not installed

    return {
        "version": __version__,
        # TODO allow for parallelization of some form
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def run_autodoc(app: Sphinx) -> None:
    """The primary sphinx call back event for sphinx."""

    # load the configuration
    values: dict[str, t.Any] = {}
    for name, _, field in Config().as_triple():
        sphinx_name = f"{CONFIG_PREFIX}{name}"
        value = app.config[sphinx_name]
        if "sphinx_validate" in field.metadata:
            try:
                value = field.metadata["sphinx_validate"](sphinx_name, value)
            except ValidationError as err:
                warn(str(err), WarningSubtypes.CONFIG_ERROR)
                continue
        values[name] = value
    config = Config(**values)

    top_level_modules = []
    for i, _ in enumerate(config.packages):
        mod_path = run_autodoc_package(app, config, i)
        if mod_path is not None:
            top_level_modules.append(mod_path)

    # create the index page
    if config.index_template:
        import jinja2

        content_str = jinja2.Template(config.index_template).render(
            top_level=top_level_modules
        )
        index_path = Path(app.srcdir) / PurePosixPath(config.output_dir) / "index.rst"
        if not (index_path.exists() and index_path.read_text("utf8") == content_str):
            index_path.parent.mkdir(parents=True, exist_ok=True)
            index_path.write_text(content_str, "utf8")


def run_autodoc_package(app: Sphinx, config: Config, pkg_index: int) -> str | None:
    """Run autodoc for a single package.

    :return: The top level module name, relative to the api directory.
    """

    package = config.packages[pkg_index]
    render_config = config.to_render_config(pkg_index)

    # find all the modules to analyse
    root_path: None | Path = Path(app.srcdir)
    if package.from_git_clone is not None:
        url, ref = package.from_git_clone
        root_path = get_git_clone(app, url, ref, config)
    if root_path is None:
        return None

    path = root_path / PurePosixPath(package.path)
    modules: t.Iterable[t.Tuple[Path, str]]
    if path.is_file():
        root_module = package.module or path.stem
        modules = [(path, root_module)]
    elif path.is_dir():
        root_module = package.module or path.name
        modules = list(
            yield_modules(
                path,
                root_module=root_module,
                exclude_dirs=config.exclude_dirs
                if package.exclude_dirs is None
                else package.exclude_dirs,
                exclude_files=config.exclude_files
                if package.exclude_files is None
                else package.exclude_files,
            )
        )
    else:
        warn(f"Package {path} does not exist", WarningSubtypes.MISSING_MODULE)
        return None

    # compute the hash of the modules,
    # so we can check if we need to re-analyse them
    hasher = hashlib.sha256()
    for mod_path, _ in sorted(modules):
        hasher.update(mod_path.read_bytes())
    hash_str = hasher.hexdigest()

    autodoc2_cache: dict[str, EnvCache]
    if not hasattr(app.env, "autodoc2_cache"):
        app.env.autodoc2_cache = autodoc2_cache = {}  # type: ignore
    else:
        autodoc2_cache = app.env.autodoc2_cache

    if (
        path.as_posix() in autodoc2_cache
        and autodoc2_cache[path.as_posix()]["hash"] == hash_str
    ):
        LOGGER.info(f"[Autodoc2] Using cached analysis {root_module!r}")
        db = autodoc2_cache[path.as_posix()]["db"]
    else:
        # analyse the modules and write to the database
        db = InMemoryDb()
        for mod_path, mod_name in status_iterator(
            modules,
            "[Autodoc2] Analysing package...",
            length=len(modules),
            stringify_func=(lambda x: str(x[1])),
        ):
            for data in analyse_module(mod_path, mod_name):
                try:
                    db.add(data)
                except UniqueError:
                    warn(
                        f"Duplicate item {data['full_name']} ({data['type']})",
                        WarningSubtypes.DUPLICATE_ITEM,
                    )
        autodoc2_cache[path.as_posix()] = {"hash": hash_str, "db": db}

    # find all the package/module, so we know what files to write
    # To following __all__ strategy
    LOGGER.info("[Autodoc2] Determining files to write ...")
    to_write: t.List[str] = []
    stack = [root_module]
    while stack:
        item = stack.pop()
        to_write.append(item)
        stack.extend(
            [
                mod
                for mod in db.get_children_names(item, {"package", "module"})
                if not any(
                    pat.fullmatch(mod) for pat in render_config.skip_module_regexes
                )
            ]
        )

    resolved_all: None | t.Dict[str, ResolvedDict] = None
    if render_config.module_all_regexes:
        LOGGER.info("[Autodoc2] Resolving __all__ imports ...")
        resolved_all = {
            k: v
            for k, v in resolve_all(db, root_module).items()
            if any(pat.fullmatch(k) for pat in render_config.module_all_regexes)
        }

    def _warn_render(msg: str, type_: WarningSubtypes) -> None:
        warn(msg, type_)

    # write the files
    output = Path(app.srcdir) / PurePosixPath(config.output_dir) / root_module
    output.mkdir(parents=True, exist_ok=True)
    paths = []
    for mod_name in status_iterator(
        to_write,
        "[Autodoc2] Writing modules...",
        length=len(to_write),
        stringify_func=(lambda x: str(x)),
    ):
        render_cls = config.render_plugin
        for pat, cls in config.render_plugin_regexes:
            if pat.fullmatch(mod_name):
                render_cls = cls
                break
        content = "\n".join(
            render_cls(db, render_config, _warn_render, resolved_all).render_item(
                mod_name
            )
        )
        out_path = output / (mod_name + render_cls.EXTENSION)
        paths.append(out_path)
        if not (out_path.exists() and out_path.read_text("utf8") == content):
            # Don't write the file if it hasn't changed
            # this means that sphinx doesn't mark it for rebuild (mtime based)
            out_path.write_text(content, "utf8")

    # clean old files
    for path in output.iterdir():
        if path.is_dir():
            continue
        if path not in paths:
            path.unlink()

    return f"{root_module}/{root_module}"


def get_git_clone(
    app: Sphinx, url: str, branch_tag: str, config: Config
) -> None | Path:
    """Download a git repository to the given folder."""
    # create a sha of the url and branch
    folder = hashlib.sha256(url.encode("utf8") + branch_tag.encode("utf8")).hexdigest()
    path = Path(app.outdir) / "_autodoc2_git_clones" / folder
    if path.exists():
        return path
    LOGGER.info(f"[Autodoc2] Cloning {url}@{branch_tag}")
    path.parent.mkdir(parents=True, exist_ok=True)
    retries = 3
    while retries:
        try:
            subprocess.check_call(
                [
                    "git",
                    "clone",
                    "--single-branch",
                    "--branch",
                    branch_tag,
                    url,
                    str(path),
                ],
                env={"GIT_TERMINAL_PROMPT": "0"},
                stdout=subprocess.DEVNULL,
            )
            break
        except subprocess.CalledProcessError as err:
            LOGGER.info(f"[Autodoc2] retrying clone {url}@{branch_tag} after: {err}")
            if path.exists():
                # remove any partial clone
                with suppress(Exception):
                    shutil.rmtree(path)
            sleep(2)
            retries -= 1
    else:
        warn(f"Failed to clone {url}@{branch_tag}", WarningSubtypes.GIT_CLONE_FAILED)
        return None
    return path


class EnvCache(t.TypedDict):
    """Cache for the environment."""

    hash: str  # the hash of the package files
    db: InMemoryDb



"""The configuration for this packages documentation."""
from datetime import date

from autodoc2 import __version__

# -- Project information -----------------------------------------------------

project = "sphinx-autodoc2"
version = __version__
copyright = f"{date.today().year}, Chris Sewell"
author = "Chris Sewell"

# -- General configuration ---------------------------------------------------

extensions = [
    "myst_parser",
    "autodoc2",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",  # for aiida
]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
}
myst_enable_extensions = ["fieldlist", "deflist"]

# -- HTML output -------------------------------------------------

html_theme = "furo"
html_title = "sphinx-autodoc2"
# html_logo = "logo.svg"
html_theme_options = {
    "top_of_page_button": "edit",
    "source_repository": "https://github.com/chrisjsewell/sphinx-autodoc2/",
    "source_branch": "main",
    "source_directory": "docs/",
}

# --- Autodoc configuration ------

autodoc2_packages = [
    "../src/autodoc2",
    {
        "path": "aiida",
        "from_git_clone": (
            "https://github.com/aiidateam/aiida-core.git",
            "v2.2.2",
        ),
        "skip_module_regexes": [
            r"aiida\.[^\.]+\..*",
            r"aiida\.(__main__|calculations|restapi|sphinxext|storage|workflows)",
        ],
        "module_all_regexes": [
            r"aiida\.[^\.]+",
        ],
        "deprecated_module_regexes": [
            r"aiida\.parsers\.parser",
        ],
        "exclude_dirs": [
            "__pycache__",
            # "migrations",
        ],
    },
]
autodoc2_render_plugin_regexes = [(r"autodoc2\.db", "myst")]
# autodoc2_sort_names = True

nitpick_ignore_regex = [
    ("py:class", r"re\.Pattern"),
    (r"py:.*", r"typing_extensions.*"),
    (r"py:.*", r"astroid.*"),
    # TODO for some reason in:
    # .. py:function:: format_args(args_info: autodoc2.utils.ARGS_TYPE ...
    # ARGS_TYPE is treated as a class rather than data
    ("py:class", r"autodoc2\.utils\..*_TYPE"),
    # from aiida
    (r"py:.*", r"types\.FunctionType"),
    (
        r"py:.*",
        r"(aiida|circus|click|disk_objectstore|importlib_metadata|kiwipy|plumpy|pgsu|requests).*",
    ),
    (r"py:.*", r"(DaemonException|Manager|PersistenceError)"),
    (r"py:.*", r"QueryBuilder\._get_ormclass"),
]

# --- Additional configuration ----

from dataclasses import fields  # noqa: E402
import typing as t  # noqa: E402

from docutils import nodes  # noqa: E402
from sphinx.application import Sphinx  # noqa: E402
from sphinx.util.docutils import SphinxDirective  # noqa: E402

from autodoc2.config import CONFIG_PREFIX, Config, PackageConfig  # noqa: E402


def setup(app: Sphinx) -> None:
    app.add_object_type(
        "confval",  # directivename
        "confval",  # rolename
        "pair: %s; configuration value",  # indextemplate
    )

    app.add_directive("autodoc2-config", CreateConfigDirective)
    app.add_directive("autodoc2-config-package", CreateConfigPkgDirective)


class CreateConfigDirective(SphinxDirective):
    """Document the configuration options."""

    def run(self) -> t.List[nodes.Node]:
        text = []

        config = Config()
        for name, value, field in config.as_triple():
            text.append(f"``````{{confval}} {CONFIG_PREFIX}{name}")
            text.append(f'{field.metadata.get("help", "")}')
            text.append("")
            if "category" in field.metadata:
                text.append(f"**category**: { field.metadata['category']}")
                text.append("")
            if "sphinx_type" in field.metadata:
                type_ = type_to_string(field.metadata["sphinx_type"])
                text.append(f"**type**: {type_}")
                text.append("")
            default_str = value if isinstance(value, str) else repr(value)
            if len(default_str.splitlines()) == 1:
                text.append(f"**default**: ``{default_str}``")
            else:
                text.append("**default**:")
                text.append("")
                text.append("```")
                text.append(default_str)
                text.append("```")
            text.append("")
            text.append("``````")

        base_node = nodes.Element()
        self.state.nested_parse(text, self.content_offset, base_node)
        return base_node.children  # type: ignore


class CreateConfigPkgDirective(SphinxDirective):
    """Document the package-level configuration options."""

    def run(self) -> t.List[nodes.Node]:
        base_node = nodes.Element()
        text = []
        pkg_config = PackageConfig("")
        config_fields = {f.name: f for f in fields(Config)}
        for name, _, field in pkg_config.as_triple():
            text.append(f"``````{{confval}} {CONFIG_PREFIX}packages[{name}]")
            if name in config_fields:
                text.append("*global override*")
                text.append("")
                text.append(f'{config_fields[name].metadata.get("help", "")}')
            else:
                text.append(f'{field.metadata.get("help", "")}')
            text.append("``````")
        self.state.nested_parse(text, self.content_offset, base_node)
        return base_node.children  # type: ignore


def type_to_string(type_: t.Type[t.Any]) -> str:
    """Convert a type to a string."""
    # TODO just keeping it simple for now but can we do this with astroid!?
    if type_ is str:
        return "``str``"
    if type_ is int:
        return "``int``"
    if type_ is bool:
        return "``bool``"
    if type_ is float:
        return "``float``"
    if type_ is list:
        return "``list``"
    type_string = str(type_)
    if type_string == "typing.Optional[str]":
        return "``str`` or ``None``"
    if "RstRender" in type_string:
        return '``"rst"``'
    return type_string



"""Convert the database items into documentation."""
# note, for the directives and options see:
# https://www.sphinx-doc.org/en/master/usage/restructuredtext/domains.html
from __future__ import annotations

import abc
from collections import OrderedDict
import typing as t

from autodoc2.utils import WarningSubtypes

if t.TYPE_CHECKING:
    from autodoc2.config import RenderConfig
    from autodoc2.db import Database
    from autodoc2.utils import ARGS_TYPE, ItemData, ResolvedDict


class RendererBase(abc.ABC):
    """The base renderer."""

    EXTENSION: t.ClassVar[str] = ".txt"
    """The extension for the output files."""

    def __init__(
        self,
        db: Database,
        config: RenderConfig,
        warn: t.Callable[[str, WarningSubtypes], None] | None = None,
        resolved_all: dict[str, ResolvedDict] | None = None,
    ) -> None:
        """Initialise the renderer."""
        self._db = db
        self._config = config
        self._warn = warn or (lambda msg, type_: None)
        self._resolved_all = resolved_all
        self._resolve_all_warned: set[str] = set()
        """The full_names of modules that have already been warned about, regarding __all__ resolution"""
        self._is_hidden_cache: OrderedDict[str, bool] = OrderedDict()
        """Cache for the is_hidden function: full_name -> bool."""

    @property
    def config(self) -> RenderConfig:
        """The configuration."""
        return self._config

    def warn(
        self, msg: str, type_: WarningSubtypes = WarningSubtypes.RENDER_ERROR
    ) -> None:
        """Warn the user."""
        self._warn(msg, type_)

    def get_item(self, full_name: str) -> t.Optional[ItemData]:
        """Get an item from the database, by full_name."""
        return self._db.get_item(full_name)

    def get_children(
        self,
        item: ItemData,
        types: None | set[str] = None,
        *,
        omit_hidden: bool = True,
    ) -> t.Iterable[ItemData]:
        """Get the children of this item, sorted according to the config.

        If module and full_name in module_all_regexes,
        it will use the __all__ list instead of the children.

        :param item: The item to get the children of.
        :param types: If given, only return items of these types.
        :param omit_hidden: If True, omit hidden items.
        """
        if item["type"] in {"module", "package"} and any(
            pat.fullmatch(item["full_name"]) for pat in self.config.module_all_regexes
        ):
            warn = item["full_name"] not in self._resolve_all_warned
            self._resolve_all_warned.add(item["full_name"])
            if item.get("all") is None and warn:
                self.warn(
                    f"__all__ missing or empty in {item['full_name']}",
                    WarningSubtypes.ALL_MISSING,
                )
            resolved_data = (
                self._resolved_all.get(item["full_name"])
                if self._resolved_all
                else None
            )
            if resolved_data is None:
                if warn:
                    self.warn(
                        f"Could not resolve {item['full_name']}.__all__",
                        WarningSubtypes.ALL_RESOLUTION,
                    )
                return
            if warn:
                for key, message in (
                    (
                        "unresolved",
                        "Could not resolve {names!r}, in {full_name}.__all__",
                    ),
                    (
                        "stars_unknown",
                        "* imports from unknown {names!r}, in {full_name}.__all__",
                    ),
                    (
                        "stars_no_all",
                        "* imports from modules with no __all__ {names!r}, in {full_name}.__all__",
                    ),
                    (
                        "stars_unresolved",
                        "* imports from modules with no resolved __all__ {names!r}, in {full_name}.__all__",
                    ),
                ):
                    names = resolved_data[key]  # type: ignore
                    if names:
                        self.warn(
                            message.format(names=names, full_name=item["full_name"]),
                            WarningSubtypes.ALL_RESOLUTION,
                        )
            for all_name in (
                sorted(item.get("all") or [])
                if self.config.sort_names
                else item.get("all") or []
            ):
                if all_name not in resolved_data["resolved"]:
                    continue
                resolved_names = resolved_data["resolved"][all_name]
                if len(resolved_names) > 1 and warn:
                    self.warn(
                        f"Found multiple items for {all_name!r} in "
                        f"{item['full_name']}.__all__: {resolved_names!r}",
                        WarningSubtypes.ALL_RESOLUTION,
                    )
                for resolved_name in resolved_names:
                    resolved_item: None | ItemData
                    if (
                        types == {"external"}
                        and resolved_name.split(".")[0]
                        != item["full_name"].split(".")[0]
                    ):
                        # special case, for use only in the summary table
                        # These are items that are not in the same module as the parent,
                        # that were exposed via __all__.
                        resolved_item = {
                            "full_name": resolved_name,
                            "type": "external",
                            "doc": "",
                        }
                    else:
                        resolved_item = self._db.get_item(resolved_name)
                    if not (
                        resolved_item is None
                        or (types is not None and resolved_item["type"] not in types)
                        or (omit_hidden and self.is_hidden(resolved_item))
                    ):
                        yield resolved_item
        else:
            for child in self._db.get_children(
                item["full_name"], types=types, sort_name=self.config.sort_names
            ):
                if omit_hidden and self.is_hidden(child):
                    continue
                yield child

    def is_hidden(self, item: ItemData) -> bool:
        """Whether this object should be displayed in documentation.

        Based on configuration regarding:

        - does i match a hidden regex pattern
        - does it have documentation
        - is it a dunder, i.e. __name__
        - is it a private member, i.e. starts with _, but not a dunder
        - is it an inherited member of a class
        """
        if item["full_name"] in self._is_hidden_cache:
            return self._is_hidden_cache[item["full_name"]]
        # TODO also whether it is imported, but this is only when following __all__
        is_hidden = (
            any(p.fullmatch(item["full_name"]) for p in self.config.hidden_regexes)
            or (
                item["type"] in ("module", "package")
                and any(
                    p.fullmatch(item["full_name"])
                    for p in self.config.skip_module_regexes
                )
            )
            or ("undoc" in self.config.hidden_objects and not item.get("doc", ""))
            or (
                "inherited" in self.config.hidden_objects
                and item.get("inherited", False)
            )
            or (
                "dunder" in self.config.hidden_objects
                and item["full_name"].startswith("__")
                and item["full_name"].endswith("__")
            )
            or (
                "private" in self.config.hidden_objects
                and item["full_name"].startswith("_")
                and not item["full_name"].endswith("__")
            )
        )
        self._is_hidden_cache[item["full_name"]] = is_hidden
        if len(self._is_hidden_cache) > 1000:
            self._is_hidden_cache.popitem(last=False)
        return is_hidden

    def is_module_deprecated(self, item: ItemData) -> bool:
        """Whether this module is deprecated."""
        return any(
            p.fullmatch(item["full_name"])
            for p in self.config.deprecated_module_regexes
        )

    # TODO allow for adding :noindex:

    def show_module_summary(self, item: ItemData) -> bool:
        """Whether to show a summary for this module/package."""
        return self.config.module_summary

    def show_class_inheritance(self, item: ItemData) -> bool:
        """Whether to show the inheritance for this class."""
        return self.config.class_inheritance

    def show_annotations(self, item: ItemData) -> bool:
        """Whether to show type annotations."""
        return self.config.annotations

    @abc.abstractmethod
    def render_item(self, full_name: str) -> t.Iterable[str]:
        """Yield the content for a single item."""


def format_args(
    args_info: ARGS_TYPE,
    include_annotations: bool = True,
    ignore_self: None | str = None,
) -> str:
    """Format the arguments of a function or method."""
    result = []

    for i, (prefix, name, annotation, default) in enumerate(args_info):
        if i == 0 and ignore_self is not None and name == ignore_self:
            continue
        formatted = (
            (prefix or "")
            + (name or "")
            + (f": {annotation}" if annotation and include_annotations else "")
            + ((" = {}" if annotation else "={}").format(default) if default else "")
        )
        result.append(formatted)

    return ", ".join(result)



"""Utility functions and types."""
from __future__ import annotations

import enum
from fnmatch import fnmatch
import itertools
import os
from pathlib import Path
import typing as t

from typing_extensions import Required

if t.TYPE_CHECKING:
    from .db import Database


PROPERTY_TYPE = t.Literal[
    "async", "staticmethod", "classmethod", "abstractmethod", "singledispatch"
]
ARGS_TYPE = t.List[
    t.Tuple[t.Optional[str], t.Optional[str], t.Optional[str], t.Optional[str]]
]


class ItemData(t.TypedDict, total=False):
    """A data item, for the results of the analysis."""

    type: Required[str]
    full_name: Required[str]  # this is delimited by dots
    doc: Required[str]

    range: tuple[int, int]

    ## module / package
    file_path: None | str
    encoding: str
    all: None | list[str]
    imports: list[tuple[str, str | None]]  # path, alias

    # assign (data)
    value: None | str | t.Any  # TODO make value JSON serializable
    annotation: None | str

    # function/method/overload
    properties: list[PROPERTY_TYPE]
    args: ARGS_TYPE
    return_annotation: None | str

    # class
    bases: list[str]

    # child of class
    inherited: bool


class WarningSubtypes(enum.Enum):
    """The subtypes of warnings for the extension."""

    CONFIG_ERROR = "config_error"
    """Issue with configuration validation."""
    GIT_CLONE_FAILED = "git_clone"
    """Failed to clone a git repository."""
    MISSING_MODULE = "missing_module"
    """If the package file/folder does not exist."""
    DUPLICATE_ITEM = "dup_item"
    """Duplicate fully qualified name found during package analysis."""
    RENDER_ERROR = "render"
    """Generic rendering error."""
    ALL_MISSING = "all_missing"
    """__all__ attribute missing or empty in a module."""
    ALL_RESOLUTION = "all_resolve"
    """Issue with resolution of an item in a module's __all__ attribute."""


def yield_modules(
    folder: str | Path,
    *,
    root_module: str | None = None,
    extensions: t.Sequence[str] = (".py", ".pyi"),
    exclude_dirs: t.Sequence[str] = ("__pycache__",),
    exclude_files: t.Sequence[str] = (),
) -> t.Iterable[tuple[Path, str]]:
    """Walk the given folder and yield all required modules.

    :param folder: The path to walk.
    :param root_module: The name of the root module,
        otherwise the folder name is used.
    :param extensions: The extensions to include.
        If multiple files with the same stem,
        only the first extension will be used.
    :param exclude_dirs: Directory names to exclude (matched with fnmatch).
    :param exclude_files: File names to exclude (matched with fnmatch).
    """
    folder = Path(folder)
    root_mod = (root_module or folder.name).split(".")
    exc_dirs = set(exclude_dirs or [])
    exc_files = set(exclude_files or [])

    def _suffix_sort_key(s: str) -> int:
        return extensions.index(s)

    for root, dirs, filenames in os.walk(folder, topdown=True):
        dirs[:] = [d for d in dirs if not any(fnmatch(d, m) for m in exc_dirs)]
        to_yield: dict[str, list[str]] = {}
        for filename in filenames:
            if any(fnmatch(filename, m) for m in exc_files):
                continue
            name, suffix = os.path.splitext(filename)
            if suffix in extensions:
                to_yield.setdefault(name, []).append(suffix)
        root_path = Path(root)
        rel_mod = root_path.relative_to(folder).parts
        for name, suffixes in to_yield.items():
            suffix = sorted(suffixes, key=_suffix_sort_key)[0]
            yield (root_path / f"{name}{suffix}", ".".join([*root_mod, *rel_mod, name]))


class ResolvedDict(t.TypedDict):
    """The outcome of resolving a packages's __all__ attributes."""

    resolved: dict[str, set[str]]
    """These are names in __all__ that we have successfully resolved,
    to one or more items.
    """
    unresolved: set[str]
    """These are names in __all__ that we have not been able to resolve."""
    stars_unresolved: set[str]
    """These are star imports in the module, that we have not yet resolved."""
    stars_no_all: set[str]
    """These are star imports in the module, that have no recognised __all__.
    At present we do not resolve them.
    """
    stars_unknown: set[str]
    """These are star imports in the module, that we cannot resolve.
    Because they point to a module that is not in the database.
    """


def resolve_all(db: Database, package_name: str) -> dict[str, ResolvedDict]:
    """Give a module name, yield all names that would be imported by star."""
    # see: https://docs.python.org/3/reference/simple_stmts.html#import

    # gather all the packages and modules in the database,
    # that are, or begin with package_name
    modules = {
        i["full_name"]: i
        for i in itertools.chain(db.get_by_type("module"), db.get_by_type("package"))
        if (
            i["full_name"].startswith(package_name + ".")
            or i["full_name"] == package_name
        )
    }

    res_data: dict[str, ResolvedDict] = {}

    # first do the "local" resolution,
    # i.e. everything that is not a star import
    for name in modules:
        module = modules[name]
        all_ = module.get("all") or []

        res_data[name] = {
            "resolved": {},
            "unresolved": set(),
            "stars_unresolved": set(),
            "stars_no_all": set(),
            "stars_unknown": set(),
        }

        if not all_:
            continue

        # first look at local children
        children = {
            n["full_name"].split(".")[-1]: n["full_name"]
            for n in db.get_children(name)
            if n["type"] not in {"module", "package"}
        }
        for all_name in all_:
            if all_name in children:
                res_data[name]["resolved"].setdefault(all_name, set()).add(
                    children[all_name]
                )

        # now look at imports
        for import_name, alias in module.get("imports", []):
            final_name = alias or import_name.split(".")[-1]
            if final_name == "*":
                if import_name[:-2] not in modules:
                    # these are unresolvable so we just record, so we can warn later
                    res_data[name]["stars_unknown"].add(import_name[:-2])
                elif modules[import_name[:-2]].get("all") is None:
                    # There is no recognised __all__ in this module
                    # TODO here we should check against all the modules children and imports
                    # to directly mirror the behaviour of the interpreter
                    # but for now we just record, so we can warn later
                    res_data[name]["stars_no_all"].add(import_name[:-2])
                else:
                    # we cannot resolve these stars yet, so we just record for now
                    res_data[name]["stars_unresolved"].add(import_name[:-2])
            elif final_name in all_:
                res_data[name]["resolved"].setdefault(final_name, set()).add(
                    import_name
                )

    # now resolve the stars, by continually iterating
    # to find/resolve modules with no pointers to other unresolved modules
    something_changed = True
    while something_changed:
        something_changed = False
        for module_name, module_data in res_data.items():
            module_all = modules[module_name].get("all") or []
            for star_name in list(module_data["stars_unresolved"]):
                if res_data[star_name]["stars_unresolved"]:
                    # we have not fully resolved this module yet
                    # so have to wait until the next iteration
                    continue
                module_data["stars_unresolved"].discard(star_name)
                something_changed = True
                for known, known_full_names in res_data[star_name]["resolved"].items():
                    if known in module_all:
                        module_data["resolved"].setdefault(known, set()).update(
                            known_full_names
                        )

    # add the final unresolved names
    for module_name, module_data in res_data.items():
        module_data["unresolved"] = set(modules[module_name]["all"] or []) - set(
            module_data["resolved"]
        )

    return res_data


# Deprecated as a bit of a hack, but leave here for now, to record
# def resolve_all_item(
#     db: Database, module_item: ItemData, all_name: str
# ) -> t.Iterable[str]:
#     """Resolve an item in an `__all__` list, to its full name(s)."""

#     module_name = module_item["full_name"]
#     base = module_name.split(".")[0]

#     item = db.get_item(f"{module_name}.{all_name}")
#     if item and item["type"] not in {"package", "module"}:
#         # directly declared in the module
#         yield f"{module_name}.{all_name}"

#     for import_name, alias in module_item.get("imports") or []:
#         if not import_name.startswith(f"{base}."):
#             continue

#         import_short_name = alias or import_name.split(".")[-1]

#         if import_short_name == all_name:
#             # directly declared in an import within the module
#             yield import_name

#         elif import_short_name == "*":
#             # imported as an ancestor of an import within the module
#             import_parent_name = import_name[:-2]
#             _child_item = db.get_item(f"{import_parent_name}.{all_name}")
#             for ancestor in itertools.chain(
#                 [] if _child_item is None else [_child_item],
#                 db.get_items_like(f"{import_parent_name}.*.{all_name}"),
#             ):
#                 ancestor_parts = ancestor["full_name"].split(".")
#                 ancestor_parent = db.get_item(".".join(ancestor_parts[:-1]))
#                 # TODO we should check that they are in all `__all__` s, up to the module
#                 if ancestor_parent and ancestor_parts[-1] in (
#                     ancestor_parent.get("all") or []
#                 ):
#                     yield ancestor["full_name"]



"""Basic tests for the analysis module."""
from __future__ import annotations

import typing as t

import pytest

from autodoc2.analysis import analyse_module


def clean_item(item: dict[str, t.Any]) -> dict[str, t.Any]:
    """Remove non-deterministic data."""
    return {
        key: value
        for key, value in item.items()
        if key not in ("modified", "file_path", "encoding")
    }


test_items = {
    # module
    "module_docstring": "'''Docstring'''",
    "module_all": "__all__ = ['a', 'b', 'c']",
    # annotation only assign
    "assign_annotation": "a: str",
    "assign_annotation_union": "a: str | int",
    # assign with annotation
    "assign_string": "a: str = 'Hello World'\n'''Docstring'''",
    "assign_bytes": "a: bytes = b'Hello World'",
    "assign_int": "a: int = 42",
    "assign_float": "a: float = 3.14",
    "assign_list": "a: list = [1, 2, 3]",
    "assign_tuple": "a: tuple = (1, 2, 3)",
    "assign_dict": "a: dict[str, int] = {'a': 1, 'b': 2, 'c': 3}",
    "assign_set": "a: set[int] = {1, 2, 3}",
    "assign_with_docstring": "a: str = 'Hello World'\n'''Docstring'''",
    # function with annotation
    "func_basic": "def a() -> str: ...",
    "func_1arg": "def a(b: int) -> str: ...",
    "func_args": "def a(*b: int) -> str: ...",
    "func_1kwarg": "def a(*, b: int) -> str: ...",
    "func_kwargs": "def a(**kw: float) -> str: ...",
    "func_docstring": "def a() -> str:\n    '''Docstring'''\n",
    "func_async": "async def a() -> str: ...",
    # class with annotation
    "class_basic": "class A: ...",
    "class_inherit": "class A(B): ...",
    "class_inherit_multi": "class A(B, C): ...",
    "class_inherit_generic": "class A(B[int]): ...",
    "class_docstring": "class A:\n    '''Docstring'''\n",
    "class_attribute": "class A:\n    a: str",
    "class_attribute_with_docstring": "class A:\n    a: str\n    '''Docstring'''\n",
    # class method with annotation
    "class_method_basic": "class A:\n    def a(self) -> str: ...",
    # class property with annotation
    "class_property_basic": "class A:\n    @property\n    def a(self) -> str: ...",
    # class static method with annotation
    "class_staticmethod_basic": "class A:\n    @staticmethod\n    def a() -> str: ...",
    # class class method with annotation
    "class_classmethod_basic": "class A:\n    @classmethod\n    def a(cls) -> str: ...",
    # class abstract method with annotation
    "class_abstractmethod_basic": "import abc\n\nclass A:\n    @abc.abstractmethod\n    def a(self) -> str: ...",
    "class_inherited": "class A:\n    def a(self) -> str: ...\nclass B(A): ...",
    "class_inherited_with_docstring": "class A:\n    '''Docstring'''\n\n    def a(self) -> str: ...\nclass B(A): ...",
    # overloads
    "overload_basic": "from typing import overload\n\n@overload\ndef a() -> str: ...\n@overload\ndef a(b: int) -> str: ...",
}


@pytest.mark.parametrize(
    "content",
    list(test_items.values()),
    ids=list(test_items),
)
def test_basic(content, tmp_path, data_regression):
    tmp_path.joinpath("test.py").write_text(content)
    data = [clean_item(item) for item in analyse_module(tmp_path / "test.py", "test")]
    data_regression.check(data)


def test_package(tmp_path, data_regression):
    pkg = tmp_path.joinpath("test")
    pkg.mkdir()
    pkg.joinpath("__init__.py").write_text("'''Docstring'''")
    pkg.joinpath("a.py").write_text("a: str = 'Hello World'")
    pkg.joinpath("b.py").write_text(
        "from .a import a\nfrom .a import a as x\nb: str = a"
    )
    data = {
        "init": [
            clean_item(item)
            for item in analyse_module(pkg.joinpath("__init__.py"), "test")
        ],
        "a": [
            clean_item(item) for item in analyse_module(pkg.joinpath("a.py"), "test.a")
        ],
        "b": [
            clean_item(item) for item in analyse_module(pkg.joinpath("b.py"), "test.b")
        ],
    }
    data_regression.check(data)



"""CLI for the package."""
from pathlib import Path
import re
import typing as t

from rich import progress as rp
from rich.console import Console
import typer

from autodoc2 import __name__ as package_name
from autodoc2 import __version__
from autodoc2.analysis import analyse_module
from autodoc2.config import Config
from autodoc2.db import InMemoryDb, UniqueError
from autodoc2.utils import WarningSubtypes, yield_modules

console = Console()


app_main = typer.Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    rich_markup_mode="rich",
    no_args_is_help=True,
)


def version_callback(value: bool) -> None:
    """Print the version and exit."""
    if value:
        console.print(f"{package_name} version: {__version__}")
        raise typer.Exit()


@app_main.callback()
def main_app(
    version: t.Optional[bool] = typer.Option(
        None,
        "-v",
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show the application version and exit.",
    ),
) -> None:
    """[underline]CLI for sphinx-autodoc2[/underline]"""


@app_main.command("list")
def list_items(
    path: Path = typer.Argument(..., exists=True, help="Path to analyse"),
    module: t.Optional[str] = typer.Option(
        None,
        "-m",
        "--module",
        help="The name of the module, otherwise it will be guessed from the path",
    ),
    inherited: bool = typer.Option(
        False, "-i", "--inherited", help="Show inherited members"
    ),
    private: bool = typer.Option(False, "-p", "--private", help="Show private members"),
    one_line: bool = typer.Option(
        False, "-o", "--one-line", help="Show only full name and type"
    ),
    filter_types_str: t.Optional[str] = typer.Option(
        None,
        "-ft",
        "--filter-types",
        help="Only show members of types (comma separated)",
    ),
    skip_types_str: str = typer.Option(
        "import_from",
        "-st",
        "--skip-types",
        help="Do not show members of types (comma separated)",
    ),
    filter_name: t.Optional[str] = typer.Option(
        None, "-fn", "--filter-name", help="Only show members with this name regex"
    ),
) -> None:
    """Analyse a python module or package and stream the results to the console."""
    # initialise filters
    skip_types = skip_types_str.split(",")
    filter_types = filter_types_str.split(",") if filter_types_str else None
    filter_name_func: t.Callable[[str], bool] = (
        re.compile(filter_name).match if filter_name else lambda _: True  # type: ignore
    )

    # gather the modules
    modules: t.Iterable[t.Tuple[Path, str]]
    if path.is_dir():
        root_module = module or path.name
        modules = yield_modules(path, root_module=root_module)
    else:
        root_module = module or path.stem
        modules = [(path, root_module)]

    for mod_path, mod_name in modules:
        for data in analyse_module(mod_path, mod_name):
            if not filter_name_func(data["full_name"]):
                continue
            if data["type"] in skip_types or (
                filter_types and data["type"] not in filter_types
            ):
                continue
            if not inherited and data.get("inherited", None):
                continue
            if not private and data["full_name"].split(".")[-1].startswith("_"):
                continue
            if one_line:
                console.print(f'{data["full_name"]} ({data["type"]})')
            else:
                console.print(data)


@app_main.command("create-db")
def create_db(
    path: Path = typer.Argument(..., exists=True, help="Path to analyse"),
    output: Path = typer.Argument("autodoc.db.json", help="File to write to"),
    module: t.Optional[str] = typer.Option(
        None,
        "-m",
        "--module",
        help="The name of the module, otherwise it will be guessed from the path",
    ),
) -> None:
    """Create a database for a python module or package."""
    # gather the modules
    modules: t.Iterable[t.Tuple[Path, str]]
    if path.is_dir():
        root_module = module or path.name
        modules = list(yield_modules(path, root_module=root_module))
    else:
        root_module = module or path.stem
        modules = [(path, root_module)]

    # analyse the modules and write to the database
    db = InMemoryDb()
    with rp.Progress(
        rp.TextColumn("[bold]Analyse[/bold]"),
        rp.BarColumn(),
        rp.TaskProgressColumn(),
        rp.TimeElapsedColumn(),
        rp.TextColumn("[progress.description]{task.description}"),
    ) as progress:
        task = progress.add_task("", total=len(modules))
        for mod_path, mod_name in modules:
            progress.update(task, description=mod_name)
            for data in analyse_module(mod_path, mod_name):
                try:
                    db.add(data)
                except UniqueError:
                    progress.console.print(
                        f"[yellow]Warning[/yellow] Duplicate item {data['full_name']} ({data['type']})"
                    )
            progress.advance(task)

    # write the database to file
    with output.open("w") as f:
        db.write(f)
    console.print(f"[green]Database written[/green]: {output}")


@app_main.command("resolve-all")
def analyse_all(
    path: Path = typer.Argument(..., exists=True, help="Path to a database file"),
    package: str = typer.Argument(
        ...,
        help="The name of the package to resolve.",
    ),
) -> None:
    """Analyse the __all__ of a module and find potential matches"""
    with path.open("r") as f:
        db = InMemoryDb.read(f)

    from autodoc2.utils import resolve_all

    data = resolve_all(db, package)

    console.print(data)

    console.print("")
    console.print("[green]Done![/green]")


@app_main.command("write")
def write(
    path: Path = typer.Argument(..., exists=True, help="Path to analyse"),
    module: t.Optional[str] = typer.Option(
        None,
        "-m",
        "--module",
        help="The name of the module, otherwise it will be guessed from the path",
    ),
    # TODO read from config file, to populate config object
    output: Path = typer.Option("_autodoc", help="Folder to write to"),
    clean: bool = typer.Option(False, "-c", "--clean", help="Remove old files"),
) -> None:
    """Create sphinx files for a python module or package."""
    # gather the module
    modules: t.Iterable[t.Tuple[Path, str]]
    if path.is_dir():
        root_module = module or path.name
        modules = list(yield_modules(path, root_module=root_module))
    else:
        root_module = module or path.stem
        modules = [(path, root_module)]

    # analyse the modules and write to the database
    db = InMemoryDb()
    with rp.Progress(
        rp.TextColumn("[bold]Analyse[/bold]"),
        rp.BarColumn(),
        rp.TaskProgressColumn(),
        rp.TimeElapsedColumn(),
        rp.TextColumn("[progress.description]{task.description}"),
    ) as progress:
        task = progress.add_task("", total=len(modules))
        for mod_path, mod_name in modules:
            progress.update(task, advance=1, description=mod_name)
            for data in analyse_module(mod_path, mod_name):
                try:
                    db.add(data)
                except UniqueError:
                    progress.console.print(
                        f"[yellow]Warning[/yellow] Duplicate item {data['full_name']} ({data['type']})"
                    )

    # find all the package/module, so we know what files to write
    # To following __all__ strategy
    console.print("Determining files to write...")
    to_write: t.List[str] = []
    stack = [root_module]
    while stack:
        item = stack.pop()
        to_write.append(item)
        stack.extend(list(db.get_children_names(item, {"package", "module"})))

    # write the files
    output.mkdir(parents=True, exist_ok=True)
    paths = []
    with rp.Progress(
        rp.TextColumn("[bold]Write[/bold]"),
        rp.BarColumn(),
        rp.TaskProgressColumn(),
        rp.TimeElapsedColumn(),
        rp.TextColumn("[progress.description]{task.description}"),
    ) as progress:
        task = progress.add_task("", total=len(to_write))

        def _warn(msg: str, type_: WarningSubtypes) -> None:
            progress.console.print(f"[yellow]Warning[/yellow] {msg} [{type_.value}]")

        config = Config()
        for mod_name in to_write:
            progress.update(task, advance=1, description=mod_name)
            content = "\n".join(
                config.render_plugin(
                    db, config.to_render_config(None), _warn
                ).render_item(mod_name)
            )
            out_path = output / (mod_name + config.render_plugin.EXTENSION)
            paths.append(out_path)
            if out_path.exists() and out_path.read_text("utf8") == content:
                # Don't write the file if it hasn't changed
                # this means that sphinx doesn't mark it for rebuild (mtime based)
                continue
            out_path.write_text(content, "utf8")

    # remove any files that are no longer needed
    if clean:
        console.print("[bold]Cleaning old files[/bold]")
        for path in output.iterdir():
            if path.is_dir():
                continue
            if path not in paths:
                path.unlink()

    console.print("[bold green]Success![/bold green]")
    console.print(f"Files written to: {output}")



"""Renderer for reStructuredText."""
from __future__ import annotations

import re
from textwrap import indent
import typing as t

from autodoc2.render.base import RendererBase, format_args

if t.TYPE_CHECKING:
    from autodoc2.utils import ItemData


class RstRenderer(RendererBase):
    """Render the documentation as reStructuredText."""

    EXTENSION = ".rst"

    def render_item(self, full_name: str) -> t.Iterable[str]:
        item = self.get_item(full_name)
        if item is None:
            raise ValueError(f"Item {full_name} does not exist")
        type_ = item["type"]
        if type_ == "package":
            yield from self.render_package(item)
        elif type_ == "module":
            yield from self.render_module(item)
        elif type_ == "function":
            yield from self.render_function(item)
        elif type_ == "class":
            yield from self.render_class(item)
        elif type_ == "exception":
            yield from self.render_exception(item)
        elif type_ == "property":
            yield from self.render_property(item)
        elif type_ == "method":
            yield from self.render_method(item)
        elif type_ == "attribute":
            yield from self.render_attribute(item)
        elif type_ == "data":
            yield from self.render_data(item)
        else:
            self.warn(f"Unknown item type {type_!r} for {full_name!r}")

    def generate_summary(self, items: list[ItemData]) -> t.Iterable[str]:
        """Generate a summary of the items."""
        for item in items:
            short_name = item["full_name"].split(".")[-1]
            summary = ""
            # get the first paragraph, allowing for the first line to be blank
            for i, line in enumerate(item["doc"].splitlines()):
                if i > 0 and not line.strip():
                    break
                summary += line.rstrip() + " "
            summary = summary.strip()
            if summary.endswith("::"):
                # corner-case in RST, where this starts a literal block
                summary = summary[:-2]
            # TODO get signature (for functions, etc), plus sphinx also runs rst.escape
            yield f"   * - :py:obj:`{short_name} <{item['full_name']}>`"
            yield f"     - {summary}"

    def render_package(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a package."""
        if self.is_hidden(item):
            yield from [":orphan:", ""]

        full_name = item["full_name"]

        line = f":py:mod:`{full_name}`"
        yield from [line, "=" * len(line), ""]

        yield f".. py:module:: {full_name}"
        if self.is_module_deprecated(item):
            yield "   :deprecated:"
        yield from ["", ""]

        if item["doc"]:
            yield from [
                "Description",
                "-----------",
                "",
                # TODO ensure any headings are at least level 3
                item["doc"],
                "",
            ]

        visible_subpackages = [
            i["full_name"] for i in self.get_children(item, {"package"})
        ]
        if visible_subpackages:
            yield from [
                "Subpackages",
                "-----------",
                "",
                ".. toctree::",
                "   :titlesonly:",
                "   :maxdepth: 3",
                "",
            ]
            for name in visible_subpackages:
                yield f"   {name}"
            yield ""

        visible_submodules = [
            i["full_name"] for i in self.get_children(item, {"module"})
        ]
        if visible_submodules:
            yield from [
                "Submodules",
                "----------",
                "",
                ".. toctree::",
                "   :titlesonly:",
                "   :maxdepth: 1",
                "",
            ]
            for name in visible_submodules:
                yield f"   {name}"
            yield ""

        visible_children = [
            i["full_name"]
            for i in self.get_children(item)
            if i["type"] not in ("package", "module")
        ]
        if not visible_children:
            return

        title = f"{item['type'].capitalize()} Contents"
        yield from [title, "-" * len(title), ""]

        if self.show_module_summary(item):
            for heading, types in [
                ("Classes", {"class"}),
                ("Functions", {"function"}),
                ("Data", {"data"}),
                ("External", {"external"}),
            ]:
                visible_items = list(self.get_children(item, types))
                if visible_items:
                    yield from [
                        heading,
                        "~" * len(heading),
                        "",
                        ".. list-table::",
                        "   :class: autosummary longtable",
                        "   :align: left",
                        "",
                    ]
                    yield from self.generate_summary(visible_items)
                    yield ""

            yield from ["API", "~~~", ""]
            for name in visible_children:
                yield from self.render_item(name)

    def render_module(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a module."""
        yield from self.render_package(item)

    def render_function(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a function."""
        if self.is_hidden(item):
            return

        short_name = item["full_name"].split(".")[-1]
        show_annotations = self.show_annotations(item)
        sig = f"{short_name}({format_args(item['args'], show_annotations)})"
        if show_annotations and item.get("return_annotation"):
            sig += f" -> {item['return_annotation']}"

        yield f".. py:function:: {sig}"
        yield f"   :canonical: {item['full_name']}"
        # TODO overloads
        if "async" in item.get("properties", []):
            yield "   :async:"
            # TODO it would also be good to highlight if singledispatch decorated,
            # or, more broadly speaking, decorated at all
        yield ""
        if item["doc"]:
            yield indent(item["doc"], "   ")
            yield ""

    def render_exception(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for an exception."""
        yield from self.render_class(item)

    def render_class(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a class."""
        if self.is_hidden(item):
            return

        short_name = item["full_name"].split(".")[-1]
        constructor = self.get_item(f"{item['full_name']}.__init__")
        sig = short_name
        if constructor and "args" in constructor:
            sig += f"({format_args(constructor['args'], self.show_annotations(item), ignore_self='self')})"

        yield f".. py:{item['type']}:: {sig}"
        yield f"   :canonical: {item['full_name']}"
        yield ""

        # TODO overloads

        if item.get("bases") and self.show_class_inheritance(item):
            yield "   Bases: " + ", ".join(
                [_reformat_cls_base_rst(b) for b in item.get("bases", [])]
            )
            yield ""

        # TODO inheritance diagram

        if item["doc"]:
            yield indent(item["doc"], "   ")
            yield ""

        for child in self.get_children(
            item, {"class", "property", "attribute", "method"}
        ):
            for line in self.render_item(child["full_name"]):
                yield indent(line, "   ")

    def render_property(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a property."""
        if self.is_hidden(item):
            return

        short_name = item["full_name"].split(".")[-1]
        yield f".. py:property:: {short_name}"
        yield f"   :canonical: {item['full_name']}"
        for prop in ("abstractmethod", "classmethod"):
            if prop in item.get("properties", []):
                yield f"   :{prop}:"
        if item.get("return_annotation"):
            yield f"   :type: {item['return_annotation']}"
        yield ""

        if item["doc"]:
            yield indent(item["doc"], "   ")
            yield ""

    def render_method(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a method."""
        if self.is_hidden(item):
            return

        short_name = item["full_name"].split(".")[-1]
        show_annotations = self.show_annotations(item)
        sig = f"{short_name}({format_args(item['args'], show_annotations, ignore_self='self')})"
        if show_annotations and item.get("return_annotation"):
            sig += f" -> {item['return_annotation']}"

        yield f".. py:method:: {sig}"
        yield f"   :canonical: {item['full_name']}"
        # TODO overloads
        # TODO collect final decorated in analysis
        for prop in ("abstractmethod", "async", "classmethod", "final", "staticmethod"):
            if prop in item.get("properties", []):
                yield f"   :{prop}:"
        yield ""

        if item["doc"]:
            yield indent(item["doc"], "   ")
            yield ""

    def render_attribute(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for an attribute."""
        yield from self.render_data(item)

    def render_data(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a data item."""
        if self.is_hidden(item):
            return

        short_name = item["full_name"].split(".")[-1]
        yield f".. py:{item['type']}:: {short_name}"
        yield f"   :canonical: {item['full_name']}"
        for prop in ("abstractmethod", "classmethod"):
            if prop in item.get("properties", []):
                yield f"   :{prop}:"
        if item.get("annotation"):
            yield f"   :type: {item['annotation']}"
        value = item.get("value")
        if isinstance(value, str):
            if len(value.splitlines()) == 1:
                if len(value) > 100:
                    value = value[:100] + "..."
                yield f"   :value: {value!r}"
            else:
                yield "   :value: <Multiline-String>"
                # TODO in sphinx-autoapi, they made a code block inside a details/summary HTML
        else:
            value = str(value).replace("\n", " ")
            if len(value) > 100:
                value = value[:100] + "..."
            yield f"   :value: {value}"

        yield ""
        if item["doc"]:
            yield indent(item["doc"], "   ")
            yield ""


_RE_DELIMS = re.compile(r"(\s*[\[\]\(\),]\s*)")


def _reformat_cls_base_rst(value: str) -> str:
    """Reformat the base of a class for RST.

    Base annotations can come in the form::

        A[B, C, D]

    which we want to reformat as::

        :py:obj:`A`\\ [\\ :py:obj:`B`\\ , :py:obj:`C`\\ , :py:obj:`D`\\ ]

    The backslash escapes are needed because of:
    https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#character-level-inline-markup-1
    """
    result = ""
    for sub_target in _RE_DELIMS.split(value.strip()):
        sub_target = sub_target.strip()
        if _RE_DELIMS.match(sub_target):
            result += f"{sub_target}"
            if sub_target.endswith(","):
                result += " "
            else:
                result += "\\ "
        elif sub_target:
            result += f":py:obj:`{sub_target}`\\ "

    # Strip off the extra "\ "
    return result[:-2]



"""Tests for the database."""
from pathlib import Path
from textwrap import dedent

from autodoc2.analysis import analyse_module
from autodoc2.db import InMemoryDb
from autodoc2.utils import resolve_all, yield_modules


def test_all_resolution(tmp_path: Path, data_regression):
    """Test __all__ resolution"""
    package = tmp_path / "package"
    package.mkdir()
    package.joinpath("__init__.py").write_text(
        dedent(
            """\
        from package.a import a1
        from package.a.c import ac1 as alias
        from .b import *
        from .d import *
        from other import *
        __all__ = ['p', 'a1', 'alias', 'unknown']
        p = 1
        """
        ),
        "utf-8",
    )
    package.joinpath("a").mkdir()
    package.joinpath("a", "__init__.py").write_text(
        dedent(
            """\
        from .c import *
        from .d import *
        __all__ = ['a1', 'ac1', 'ad1', 'ade1', 'adf1']
        a1 = 1
        """
        ),
        "utf-8",
    )
    package.joinpath("a", "c.py").write_text(
        dedent(
            """\
        __all__ = ['ac1']
        ac1 = 1
        """
        ),
        "utf-8",
    )
    package.joinpath("b.py").touch()
    package.joinpath("d.py").write_text(
        # circular import
        dedent(
            """\
        from package import *
        __all__ = ['p']
        """
        ),
        "utf-8",
    )

    db = InMemoryDb()
    for path, modname in yield_modules(package):
        for item in analyse_module(path, modname):
            db.add(item)
    data_regression.check(resolve_all(db, "package"))



"""Tests for the rendering."""
import io
from pathlib import Path
from textwrap import dedent

import pytest
from sphinx.testing.util import SphinxTestApp
from sphinx.testing.util import path as sphinx_path

from autodoc2.analysis import analyse_module
from autodoc2.config import Config
from autodoc2.db import InMemoryDb
from autodoc2.render.base import RendererBase
from autodoc2.render.myst_ import MystRenderer
from autodoc2.render.rst_ import RstRenderer
from autodoc2.utils import yield_modules


@pytest.mark.parametrize(
    "renderer,extension",
    [
        (RstRenderer, ".rst"),
        (MystRenderer, ".md"),
    ],
    ids=["rst", "myst"],
)
def test_basic(renderer: RendererBase, extension: str, tmp_path: Path, file_regression):
    """Test basic rendering."""
    package = build_package(tmp_path)
    db = InMemoryDb()
    for path, modname in yield_modules(package):
        for item in analyse_module(path, modname):
            db.add(item)
    content = "\n".join(renderer(db, Config()).render_item(package.name))
    file_regression.check(content, extension=extension)


@pytest.mark.parametrize(
    "with_rebuild",
    [True, False],
    ids=["with_rebuild", "without_rebuild"],
)
def test_sphinx_build(tmp_path: Path, with_rebuild: bool):
    """Test building the Sphinx docs."""
    build_package(tmp_path)
    source = tmp_path / "source"
    source.mkdir()
    source.joinpath("index.rst").write_text(
        dedent(
            """\
        Test
        ====

        .. toctree::

            apidocs/index
        """
        ),
        "utf-8",
    )
    source.joinpath("conf.py").write_text(
        dedent(
            """\
        project = "tester"
        extensions = ["autodoc2"]
        autodoc2_packages = ["../package"]
        """
        ),
        "utf-8",
    )
    warnings = io.StringIO()
    build = tmp_path / "build"
    app = SphinxTestApp(
        buildername="html",
        srcdir=sphinx_path(source),
        builddir=sphinx_path(build),
        warning=warnings,
    )
    try:
        app.build()
    finally:
        app.cleanup()

    assertions = {}  # catch all the assertion failures, before failing the test
    if warnings.getvalue():
        assertions["warnings"] = warnings.getvalue()
    # print([p.relative_to(tmp_path).as_posix() for p in tmp_path.glob("**/*")])
    package_source = source / "apidocs" / "package" / "package.rst"
    if not (source / "apidocs" / "package" / "package.rst").exists():
        assertions["source/apidocs/package/package.rst"] = "not found"
    else:
        package_source_mtime = package_source.stat().st_mtime
    if not (build / "html" / "index.html").exists():
        assertions["build/index.html"] = "not found"
    if not (build / "html" / "apidocs" / "index.html").exists():
        assertions["build/apidocs/index.html"] = "not found"
    package_html = build / "html" / "apidocs" / "package" / "package.html"
    if not package_html.exists():
        assertions["build/apidocs/package/package.html"] = "not found"
    else:
        package_html_mtime = package_html.stat().st_mtime

    if not hasattr(app.env, "autodoc2_cache"):
        assertions["autodoc2_cache"] = "not found"
    if not getattr(app.env, "autodoc2_cache"):  # noqa: B009
        assertions["autodoc2_cache"] = "empty"

    if assertions:
        pytest.fail(f"Failed assertions {assertions}")

    if not with_rebuild:
        return

    # test rebuild
    rebuild_warnings = io.StringIO()
    rebuild_app = SphinxTestApp(
        buildername="html",
        srcdir=sphinx_path(source),
        builddir=sphinx_path(build),
        warning=rebuild_warnings,
    )
    try:
        rebuild_app.build()
    finally:
        rebuild_app.cleanup()

    if rebuild_warnings.getvalue():
        pytest.fail(f"Warnings on rebuild: {rebuild_warnings.getvalue()}")
    if package_source.stat().st_mtime != package_source_mtime:
        pytest.fail("Rebuild did not use cached source")
    if package_html.stat().st_mtime != package_html_mtime:
        pytest.fail("Rebuild did not use cached html")


def build_package(tmp_path: Path) -> Path:
    """Build a simple package for testing."""
    package = tmp_path / "package"
    package.mkdir()
    package.joinpath("__init__.py").write_text(
        dedent(
            """\
        '''This is a test package.'''
        from package.a import a1
        from package.a.c import ac1 as alias
        __all__ = ['p', 'a1', 'alias']
        p = 1
        '''p can be documented here.'''

        def func(a: str, b: int) -> alias:
            '''This is a function.'''

        class Class:
            '''This is a class.'''

            x: int = 1
            '''x can be documented here.'''

            def method(self, a: str, b: int) -> ...:
                '''This is a method.'''

            @property
            def prop(self) -> alias | None:
                '''This is a property.'''

            class Nested:
                '''This is a nested class.'''
        """
        ),
        "utf-8",
    )
    package.joinpath("a").mkdir()
    package.joinpath("a", "__init__.py").write_text(
        dedent(
            """\
        '''This is a test module.'''
        from .c import *
        from .d import *
        __all__ = ['a1', 'ac1', 'ad1', 'ade1', 'adf1']
        a1 = 1
        """
        ),
        "utf-8",
    )
    package.joinpath("a", "c.py").write_text(
        dedent(
            """\
        '''This is also a test module.'''
        __all__ = ['ac1']
        ac1 = 1
        """
        ),
        "utf-8",
    )
    return package



"""A database interface for storing and querying the analysis items."""
from __future__ import annotations

import json
import re
import typing as t

if t.TYPE_CHECKING:
    from .utils import ItemData


class UniqueError(KeyError):
    """An error raised when a unique constraint is violated."""


class Database(t.Protocol):
    """A simple interface for storing and querying the analysis items, from a single package.

    This allows for potential extensibility in the future,
    e.g. using a persistent sqlite database.
    """

    def add(self, item: ItemData) -> None:
        """Add an item to the database."""

    def __contains__(self, full_name: str) -> bool:
        """Check if an item is in the database, by full_name."""

    def get_item(self, full_name: str) -> t.Optional[ItemData]:
        """Get an item from the database, by full_name."""

    def get_items_like(self, full_name: str) -> t.Iterable[ItemData]:
        """Get an item from the database, matching the wildcards `*` and `?`.

        `*` matches any number of characters, and `?` matches any single character.
        """

    def get_type(self, full_name: str) -> None | str:
        """Get the type of an item from the database, by full_name."""

    def get_by_type(self, type_: str) -> t.Iterable[ItemData]:
        """Get all items from the database, by type."""

    def get_overloads(self, full_name: str) -> t.Iterable[ItemData]:
        """Get all function overloads for this name."""

    def get_children(
        self, full_name: str, types: None | set[str] = None, *, sort_name: bool = False
    ) -> t.Iterable[ItemData]:
        """Get all items that are direct children of this name, i.e. `{full_name}.{name}`.

        :param full_name: The full name of the item.
        :param types: If given, only return items of these types.
        :param sort_name: If True, sort the names alphabetically.
        """

    def get_children_names(
        self, full_name: str, types: None | set[str] = None, *, sort_name: bool = False
    ) -> t.Iterable[str]:
        """Get all names of direct children of this name, i.e. `{full_name}.{name}`.

        :param full_name: The full name of the item.
        :param types: If given, only return items of these types.
        :param sort_name: If True, sort the names alphabetically.
        """


_LIKE_REGEX = re.compile(r"([\*\?])")


class InMemoryDb(Database):
    """A simple in-memory database for storing and querying the analysis items."""

    def __init__(self) -> None:
        """Create the database."""
        self._items: t.Dict[str, ItemData] = {}
        self._overloads: t.Dict[str, t.List[ItemData]] = {}

    def add(self, item: ItemData) -> None:
        if item["type"] == "overload":
            # note we do this here and not in the analyser,
            # because overloads come before the function they overload
            # and we don't want the analyzer to have to "look ahead"
            self._overloads.setdefault(item["full_name"], []).append(item)
            return
        if item["full_name"] in self._items:
            raise UniqueError(f"Item {item['full_name']} already exists")
        self._items[item["full_name"]] = item

    def __contains__(self, full_name: str) -> bool:
        return full_name in self._items

    def get_item(self, full_name: str) -> t.Optional[ItemData]:
        return self._items.get(full_name)

    def get_items_like(self, full_name: str) -> t.Iterable[ItemData]:
        parts = _LIKE_REGEX.split(full_name)
        pattern = re.compile(
            "".join(
                [
                    ".*" if part == "*" else ("." if part == "?" else re.escape(part))
                    for part in parts
                ]
            )
        )
        return (
            item
            for item in self._items.values()
            if pattern.fullmatch(item["full_name"])
        )

    def get_type(self, full_name: str) -> None | str:
        item = self._items.get(full_name)
        if item is None:
            return None
        return item["type"]

    def get_by_type(self, type_: str) -> t.Iterable[ItemData]:
        return (item for item in self._items.values() if item["type"] == type_)

    def get_overloads(self, full_name: str) -> t.Iterable[ItemData]:
        return self._overloads.get(full_name, [])

    def get_children(
        self, full_name: str, types: None | set[str] = None, *, sort_name: bool = False
    ) -> t.Iterable[ItemData]:
        generator = (
            item
            for item in self._items.values()
            if item["full_name"].startswith(full_name + ".")
            and "." not in item["full_name"][len(full_name) + 1 :]
            and ((types is None) or (item["type"] in types))
        )
        if sort_name:
            return sorted(generator, key=lambda item: item["full_name"])
        return generator

    def get_children_names(
        self, full_name: str, types: None | set[str] = None, *, sort_name: bool = False
    ) -> t.Iterable[str]:
        generator = (
            item["full_name"]
            for item in self._items.values()
            if item["full_name"].startswith(full_name + ".")
            and "." not in item["full_name"][len(full_name) + 1 :]
            and ((types is None) or (item["type"] in types))
        )
        if sort_name:
            return sorted(generator)
        return generator

    def write(self, stream: t.TextIO) -> None:
        """Write the database to a file."""
        json.dump({"items": self._items, "overloads": self._overloads}, stream)

    @classmethod
    def read(cls, stream: t.TextIO) -> InMemoryDb:
        """Read the database from a file."""
        items = json.load(stream)
        db = cls()
        db._items = items["items"]
        db._overloads = items["overloads"]
        return db



"""The configuration for the extension."""
from __future__ import annotations

from copy import deepcopy
import dataclasses as dc
from importlib import import_module
import re
import typing as t

if t.TYPE_CHECKING:
    from autodoc2.render.base import RendererBase


class ValidationError(Exception):
    """An error validating a config value."""


CONFIG_PREFIX = "autodoc2_"


@dc.dataclass
class RenderConfig:
    """The configuration for rendering.

    This uses the global, with package level overrides.
    """

    module_all_regexes: list[re.Pattern[str]]
    skip_module_regexes: list[re.Pattern[str]]
    hidden_objects: set[t.Literal["undoc", "dunder", "private", "inherited"]]
    hidden_regexes: list[re.Pattern[str]]
    deprecated_module_regexes: list[re.Pattern[str]]
    module_summary: bool
    class_inheritance: bool
    annotations: bool
    sort_names: bool


@dc.dataclass
class PackageConfig:
    """A package-level config item."""

    # package specific
    path: str = dc.field(
        metadata={
            "help": "The path to the package, "
            "relative to the source directory (POSIX format)."
        }
    )
    from_git_clone: tuple[str, str] | None = dc.field(
        default=None,
        metadata={
            "help": "Clone a git (url, branch/tag)/ "
            "If using this option the 'path' will be relative to root of the cloned repository."
        },
    )
    module: str | None = dc.field(
        default=None,
        metadata={
            "help": "The module to use as the root of the package, "
            "otherwise the file/directory name is used."
        },
    )

    # global overrides
    exclude_dirs: list[str] | None = None
    exclude_files: list[str] | None = None
    module_all_regexes: list[re.Pattern[str]] | None = None
    skip_module_regexes: list[re.Pattern[str]] | None = None
    hidden_objects: set[
        t.Literal["undoc", "dunder", "private", "inherited"]
    ] | None = None
    hidden_regexes: list[re.Pattern[str]] | None = None
    deprecated_module_regexes: list[re.Pattern[str]] | None = None
    module_summary: bool | None = None
    class_inheritance: bool | None = None
    annotations: bool | None = None
    sort_names: bool | None = None

    def as_triple(self) -> t.Iterable[tuple[str, t.Any, dc.Field]]:  # type: ignore[type-arg]
        """Yield triples of (name, value, field)."""
        fields = {f.name: f for f in dc.fields(self.__class__)}
        for name, value in dc.asdict(self).items():
            yield name, value, fields[name]


def _coerce_packages(name: str, item: t.Any) -> list[PackageConfig]:
    """Coerce the packages config option to a set."""
    if not isinstance(item, (list, tuple)):
        raise ValidationError(f"{name} must be a list")
    if len(item) == 0:
        raise ValidationError(f"{name} must not be empty")
    # make sure we don't mutate the original (triggers sphinx total rebuild)
    new = [deepcopy(i) for i in item]
    for i, package in enumerate(new[:]):
        if isinstance(package, str):
            new[i] = package = {"path": package}
        if not isinstance(package, dict):
            raise ValidationError(f"{name}[{i}] must be a string or dict")
        if not isinstance(package["path"], str):
            raise ValidationError(f"{name}[{i}]['path'] must be a string")
        if package["path"].startswith("/"):
            raise ValidationError(f"{name}[{i}]['path'] must be relative")
        if "path" not in package:
            raise ValidationError(f"{name}[{i}] must have a 'path' key")
        if "from_git_clone" in package:
            if not isinstance(package["from_git_clone"], (tuple, list)):
                raise ValidationError(
                    f"{name}[{i}]['from_git_clone'] must be a tuple/list"
                )
            if len(package["from_git_clone"]) != 2:
                raise ValidationError(
                    f"{name}[{i}]['from_git_clone'] must be a tuple of length 2"
                )
            for idx in (0, 1):
                if not isinstance(package["from_git_clone"][idx], str):
                    raise ValidationError(
                        f"{name}[{i}]['from_git_clone'][{idx}] must be a string"
                    )
        if "module" in package and not isinstance(package["module"], str):
            raise ValidationError(f"{name}[{i}]['module'] must be a string")
        for key in ("exclude_files", "exclude_dirs"):
            if key in package and (
                not isinstance(package[key], list)
                or not all(isinstance(x, str) for x in package[key])
            ):
                raise ValidationError(f"{name}[{i}][{key!r}] must be a list of strings")
        for key in (
            "module_summary",
            "class_inheritance",
            "annotations",
            "sort_names",
        ):
            if key in package and not isinstance(package[key], bool):
                raise ValidationError(f"{name}[{i}][{key!r}] must be a boolean")
        for key in (
            "skip_module_regexes",
            "hidden_regexes",
            "deprecated_module_regexes",
            "module_all_regexes",
        ):
            if key in package:
                package[key] = _validate_regex_list(
                    f"{name}[{i}][{key!r}]", package[key]
                )
        if "hidden_objects" in package:
            package["hidden_objects"] = _validate_hidden_objects(
                f"{name}[{i}]['hidden_objects']", package["hidden_objects"]
            )

    return [PackageConfig(**p) for p in new]


def _validate_string_list(name: str, item: t.Any) -> list[str]:
    """Validate that an item is a string."""
    if not isinstance(item, list) or not all(isinstance(x, str) for x in item):
        raise ValidationError(f"{name!r} must be a list of string")
    return item


def _validate_hidden_objects(name: str, item: t.Any) -> set[str]:
    """Validate that the hidden objects config option is a set."""
    if not isinstance(item, (list, tuple, set)) or not all(
        isinstance(x, str) for x in item
    ):
        raise ValidationError(f"{name!r} must be a list of string")
    value = set(item)
    _valid = {"undoc", "dunder", "private", "inherited"}
    if not value.issubset(_valid):
        raise ValidationError(f"{name!r} must be a subset of {_valid}")
    return value


def _validate_regex_list(name: str, item: t.Any) -> list[t.Pattern[str]]:
    """Validate that an item is a list of regexes."""
    if not isinstance(item, list) or not all(isinstance(x, str) for x in item):
        raise ValidationError(f"{name!r} must be a list of string")
    compiled = []
    for i, regex in enumerate(set(item)):
        try:
            compiled.append(re.compile(regex))
        except re.error as exc:
            raise ValidationError(f"{name}[{i}] is not a valid regex: {exc}")
    return compiled


def _load_renderer(name: str, item: t.Any) -> type[RendererBase]:
    """Load a renderer class."""
    from autodoc2.render.base import RendererBase

    if isinstance(item, type) and issubclass(item, RendererBase):
        return item
    if not isinstance(item, str):
        raise ValidationError(
            f"{name!r} must be a string or subclass of {RendererBase.__qualname__}"
        )
    if item in ("rst", "restructuredtext"):
        from autodoc2.render.rst_ import RstRenderer

        return RstRenderer
    if item in ("markdown", "md", "myst"):
        from autodoc2.render.myst_ import MystRenderer

        return MystRenderer
    try:
        module_path, klass_name = item.rsplit(".", 1)
        mod = import_module(module_path)
        kls = getattr(mod, klass_name)
    except ImportError as exc:
        raise ValidationError(f"{name!r} could not be loaded: {exc}") from exc

    if not isinstance(kls, type) or not issubclass(kls, RendererBase):
        raise ValidationError(
            f"{name!r} must load a subclass of {RendererBase.__qualname__}"
        )
    return kls


def _load_regex_renderers(
    name: str, item: t.Any
) -> list[tuple[t.Pattern[str], type[RendererBase]]]:
    """Load a list of (regex, renderer)."""
    if not isinstance(item, (list, tuple)):
        raise ValidationError(f"{name!r} must be a list or tuple")
    new: list[tuple[t.Pattern[str], type[RendererBase]]] = []
    for i, child in enumerate(item):
        if not (isinstance(child, (list, tuple)) and len(child) == 2):
            raise ValidationError(f"{name}[{i}] must be a list/tuple of length 2")
        try:
            pattern = re.compile(child[0])
        except re.error as exc:
            raise ValidationError(f"{name}[{i}][0] is not a valid regex: {exc}")
        klass = _load_renderer(f"{name}[{i}][1]", child[1])
        new.append((pattern, klass))

    return new


@dc.dataclass
class Config:
    """The configuration for autoapi."""

    packages: list[PackageConfig] = dc.field(
        default_factory=list,
        metadata={
            "help": (
                "The packages to document. "
                "Each item can be a simple string, "
                "pointing to the package path, "
                "relative to the source directory (in POSIX format), "
                "or it can be a dictionary with more fine-grained control "
                "(see {ref}`config:package`))."
            ),
            "sphinx_type": list,
            "sphinx_validate": _coerce_packages,
            "category": "required",
        },
    )

    output_dir: str = dc.field(
        default="apidocs",
        metadata={
            "help": (
                "The root output directory for the documentation, "
                "relative to the source directory (in POSIX format)."
            ),
            "sphinx_type": str,
            "category": "render",
        },
    )

    exclude_dirs: list[str] = dc.field(
        default_factory=lambda: ["__pycache__"],
        metadata={
            "help": "Directories to exclude from module analysis (matched by fnmatch).",
            "sphinx_type": list,
            "sphinx_validate": _validate_string_list,
            "category": "analysis",
        },
    )

    exclude_files: list[str] = dc.field(
        default_factory=list,
        metadata={
            "help": "Files to exclude from module gathering (matched by fnmatch).",
            "sphinx_type": list,
            "sphinx_validate": _validate_string_list,
            "category": "analysis",
        },
    )

    render_plugin: type[RendererBase] = dc.field(
        default_factory=(lambda: _load_renderer("render_plugin", "rst")),
        metadata={
            "help": (
                "The renderer to use for the documentation. "
                "This can be one of `rst` or `md`/`myst`, "
                "to use the built-in renderers, "
                "or a string pointing to a class that inherits from `RendererBase`, "
                "such as `mypackage.mymodule.MyRenderer`."
            ),
            "sphinx_type": str,
            "sphinx_default": "rst",
            "sphinx_validate": _load_renderer,
            "category": "render",
        },
    )

    render_plugin_regexes: list[tuple[t.Pattern[str], type[RendererBase]]] = dc.field(
        default_factory=list,
        metadata={
            "help": "A list of (regex, renderer) to use for specific modules",
            "sphinx_type": list,
            "sphinx_validate": _load_regex_renderers,
            "category": "render",
        },
    )

    module_all_regexes: list[t.Pattern[str]] = dc.field(
        default_factory=list,
        metadata={
            "help": "Whether to use the `__all__` in a module, "
            "to determine what children to document",
            "sphinx_type": list,
            "sphinx_validate": _validate_regex_list,
            "category": "render",
        },
    )

    skip_module_regexes: list[t.Pattern[str]] = dc.field(
        default_factory=list,
        metadata={
            "help": "Regexes which match against module/package names, to skip them",
            "sphinx_type": list,
            "sphinx_validate": _validate_regex_list,
            "category": "render",
        },
    )

    hidden_objects: set[
        t.Literal["undoc", "dunder", "private", "inherited"]
    ] = dc.field(
        default_factory=lambda: {"inherited"},
        metadata={
            "help": (
                "The default hidden items. "
                "Can contain:\n"
                "- `undoc`: undocumented objects\n"
                "- `dunder`: double-underscore methods, e.g. `__str__`\n"
                "- `private`: single-underscore methods, e.g. `_private`\n"
                "- `inherited`: inherited class methods\n"
            ),
            "sphinx_type": list,
            "sphinx_validate": _validate_hidden_objects,
            "category": "render",
        },
    )

    hidden_regexes: list[t.Pattern[str]] = dc.field(
        default_factory=list,
        metadata={
            "help": "Regexes which match against object names, to mark them as hidden",
            "sphinx_type": list,
            "sphinx_validate": _validate_regex_list,
            "category": "render",
        },
    )

    deprecated_module_regexes: list[t.Pattern[str]] = dc.field(
        default_factory=list,
        metadata={
            "help": "Regexes which match against module names, to mark them as deprecated",
            "sphinx_type": list,
            "sphinx_validate": _validate_regex_list,
            "category": "render",
        },
    )

    module_summary: bool = dc.field(
        default=True,
        metadata={
            "help": "Whether to include a per-module summary.",
            "sphinx_type": bool,
            "category": "render",
        },
    )

    class_inheritance: bool = dc.field(
        default=True,
        metadata={
            "help": "Whether to document class inheritance.",
            "sphinx_type": bool,
            "category": "render",
        },
    )

    annotations: bool = dc.field(
        default=True,
        metadata={
            "help": "Whether to include annotations.",
            "sphinx_type": bool,
            "category": "render",
        },
    )

    sort_names: bool = dc.field(
        default=False,
        metadata={
            "help": "Whether to sort by name, when documenting, otherwise order by source",
            "sphinx_type": bool,
            "category": "render",
        },
    )

    index_template: str | None = dc.field(
        default=(
            "API Reference\n"
            "=============\n"
            "\n"
            "This page contains auto-generated API reference documentation [#f1]_.\n"
            "\n"
            ".. toctree::\n"
            "   :titlesonly:\n"
            "{% for package in top_level %}\n"
            "   {{ package }}\n"
            "{%- endfor %}\n"
            "\n"
            ".. [#f1] Created with `sphinx-autodoc2`\n"
            "\n"
        ),
        metadata={
            "help": (
                "The Jinja template for the top-level {output_dir}/index.rst, "
                "or None if no index should be written. "
                "The template will be passed a ``top_level`` variable, "
                "which is a list of top-level package/module names."
            ),
            "sphinx_type": t.Optional[str],
            "category": "render",
        },
    )

    def as_triple(self) -> t.Iterable[tuple[str, t.Any, dc.Field]]:  # type: ignore[type-arg]
        """Yield triples of (name, value, field)."""
        fields = {f.name: f for f in dc.fields(self.__class__)}
        for name, value in dc.asdict(self).items():
            yield name, value, fields[name]

    def to_render_config(self, pkg_index: int | None) -> RenderConfig:
        """Convert a module level render config."""
        # TODO config could also be specific to the module
        if pkg_index is None:
            return RenderConfig(
                skip_module_regexes=self.skip_module_regexes,
                hidden_objects=self.hidden_objects,
                hidden_regexes=self.hidden_regexes,
                deprecated_module_regexes=self.deprecated_module_regexes,
                module_summary=self.module_summary,
                class_inheritance=self.class_inheritance,
                annotations=self.annotations,
                sort_names=self.sort_names,
                module_all_regexes=self.module_all_regexes,
            )
        pkg = self.packages[pkg_index]
        return RenderConfig(
            skip_module_regexes=self.skip_module_regexes
            if pkg.skip_module_regexes is None
            else pkg.skip_module_regexes,
            hidden_objects=self.hidden_objects
            if pkg.hidden_objects is None
            else pkg.hidden_objects,
            hidden_regexes=self.hidden_regexes
            if pkg.hidden_regexes is None
            else pkg.hidden_regexes,
            deprecated_module_regexes=self.deprecated_module_regexes
            if pkg.deprecated_module_regexes is None
            else pkg.deprecated_module_regexes,
            module_summary=self.module_summary
            if pkg.module_summary is None
            else pkg.module_summary,
            class_inheritance=self.class_inheritance
            if pkg.class_inheritance is None
            else pkg.class_inheritance,
            annotations=self.annotations
            if pkg.annotations is None
            else pkg.annotations,
            sort_names=self.sort_names if pkg.sort_names is None else pkg.sort_names,
            module_all_regexes=self.module_all_regexes
            if pkg.module_all_regexes is None
            else pkg.module_all_regexes,
        )



"""Utilities for working with astroid nodes.
"""
from __future__ import annotations

import builtins
import itertools
import re
import typing as t

import astroid
from astroid import nodes


def resolve_import_alias(name: str, import_names: list[tuple[str, str | None]]) -> str:
    """Resolve a name from an aliased import to its original name.

    :param name: The potentially aliased name to resolve.
    :param import_names: The pairs of original names and aliases
        from the import.

    :returns: The original name.
    """
    resolved_name = name

    for import_name, imported_as in import_names:
        if import_name == name:
            break
        if imported_as == name:
            resolved_name = import_name
            break

    return resolved_name


def is_constructor(node: nodes.NodeNG) -> bool:
    """Check if the function is a constructor."""
    return (
        node.parent
        and isinstance(node.parent.scope(), nodes.ClassDef)
        and node.name == "__init__"
    )


def get_full_import_name(import_from: nodes.ImportFrom, name: str) -> str:
    """Get the full path of a name from a ``from x import y`` statement.

    :returns: The full import path of the name.
    """
    partial_basename = resolve_import_alias(name, import_from.names)

    module_name = import_from.modname
    if import_from.level:
        module = import_from.root()
        assert isinstance(module, nodes.Module)
        module_name = module.relative_to_absolute_name(
            import_from.modname, level=import_from.level
        )

    return f"{module_name}.{partial_basename}"


def get_assign_value(node: nodes.NodeNG) -> None | tuple[str, t.Any]:
    """Get the name and value of the assignment of the given node.

    Assignments to multiple names are ignored, as per PEP 257.

    :param node: The node to get the assignment value from.

    :returns: The name that is assigned to,
        and the value assigned to the name (if it can be converted).
    """
    try:
        targets = node.targets
    except AttributeError:
        targets = [node.target]

    if len(targets) == 1:
        target = targets[0]
        if isinstance(target, nodes.AssignName):
            name = target.name
        elif isinstance(target, nodes.AssignAttr):
            name = target.attrname
        else:
            return None
        return (name, get_const_values(node.value))

    return None


def get_const_values(node: nodes.NodeNG) -> t.Any:
    """Get the value of a constant node."""
    # TODO its not ideal that this goes to None if not understood
    # TODO better typing
    value: t.Any = None

    if isinstance(node, (nodes.List, nodes.Tuple)):
        new_value = []
        for element in node.elts:
            if isinstance(element, nodes.Const):
                new_value.append(element.value)
            elif isinstance(element, (nodes.List, nodes.Tuple)):
                new_value.append(get_const_values(element))
            else:
                break
        else:
            value = new_value

        if isinstance(node, nodes.Tuple):
            value = tuple(new_value)
    elif isinstance(node, nodes.Const):
        value = node.value

    return value


def get_assign_annotation(node: nodes.Assign) -> None | str:
    """Get the type annotation of the assignment of the given node.

    :returns: The type annotation as a string, or None if one does not exist.
    """
    annotation_node = None
    try:
        annotation_node = node.annotation
    except AttributeError:
        annotation_node = node.type_annotation

    if annotation_node is None:
        return None

    return resolve_annotation(annotation_node)


def resolve_annotation(annotation: nodes.NodeNG) -> str:
    """Resolve a type annotation to a string."""
    resolved: str
    if isinstance(annotation, nodes.Const):
        resolved = resolve_qualname(annotation, str(annotation.value))
    elif isinstance(annotation, nodes.Name):
        resolved = resolve_qualname(annotation, annotation.name)
    elif isinstance(annotation, nodes.Attribute):
        resolved = resolve_qualname(annotation, annotation.as_string())
    elif isinstance(annotation, nodes.Subscript):
        value = resolve_annotation(annotation.value)
        slice_node = annotation.slice
        if isinstance(slice_node, nodes.Index):
            slice_node = slice_node.value
        if isinstance(slice_node, nodes.Tuple):
            slice_ = ", ".join(resolve_annotation(elt) for elt in slice_node.elts)
        else:
            slice_ = resolve_annotation(slice_node)
        resolved = f"{value}[{slice_}]"
    elif isinstance(annotation, nodes.Tuple):
        resolved = (
            "(" + ", ".join(resolve_annotation(elt) for elt in annotation.elts) + ")"
        )
    elif isinstance(annotation, nodes.List):
        resolved = (
            "[" + ", ".join(resolve_annotation(elt) for elt in annotation.elts) + "]"
        )
    elif isinstance(annotation, nodes.BinOp):
        resolved = (
            resolve_annotation(annotation.left)
            + " "
            + annotation.op
            + " "
            + resolve_annotation(annotation.right)
        )
    else:
        resolved = annotation.as_string()

    # note sphinx-autoapi had this, but
    # (a) its not needed, because sphinx strips it in the rendered HTML
    # and (b) it could lead to incorrect links for name clashes

    # if resolved.startswith("typing."):
    #     return resolved[len("typing.") :]

    # Note, sphinx-autoapi had this, with the rationale:
    # > Sphinx is capable of linking anything in the same module
    # > without needing a fully qualified path.
    # However, (a) this breaks if following __all__ and
    # (b) lead to hard to decipher missing references
    # this does though lead to the fully qualified names
    # showing in the type annotations of the output HTML
    # TODO make sphinx not show fully qualified names of type annotations

    # module_prefix = annotation.root().name + "."
    # if resolved.startswith(module_prefix):
    #     return resolved[len(module_prefix) :]

    # TODO sphinx type resolver does not understand Ellipsis, maybe it should?
    if resolved == "Ellipsis":
        return "..."

    return resolved


def resolve_qualname(node: nodes.NodeNG, basename: str) -> str:
    """Resolve where a node is defined to get its fully qualified name.

    :param node: The node representing the base name.
    :param basename: The partial base name to resolve.

    :returns: The fully resolved base name.
    """
    full_basename = basename

    top_level_name = re.sub(r"\(.*\)", "", basename).split(".", 1)[0]
    # Disable until pylint uses astroid 2.7
    lookup_node = (
        node if isinstance(node, nodes.node_classes.LookupMixIn) else node.scope()
    )

    assigns = lookup_node.lookup(top_level_name)[1]

    for assignment in assigns:
        if isinstance(assignment, nodes.ImportFrom):
            import_name = get_full_import_name(assignment, top_level_name)
            full_basename = basename.replace(top_level_name, import_name, 1)
            break
        if isinstance(assignment, nodes.Import):
            import_name = resolve_import_alias(top_level_name, assignment.names)
            full_basename = basename.replace(top_level_name, import_name, 1)
            break
        if isinstance(assignment, nodes.ClassDef):
            full_basename = assignment.qname()
            break
        if isinstance(assignment, nodes.AssignName):
            full_basename = f"{assignment.scope().qname()}.{assignment.name}"

    if isinstance(node, nodes.Call):
        full_basename = re.sub(r"\(.*\)", "()", full_basename)

    if full_basename.startswith("builtins."):
        return full_basename[len("builtins.") :]

    if full_basename.startswith("__builtin__."):
        return full_basename[len("__builtin__.") :]

    return full_basename


def get_module_all(node: nodes.Module) -> None | list[str]:
    """Get the contents of the ``__all__`` variable from a module."""
    all_ = None

    if "__all__" in node.locals:
        assigned = next(node.igetattr("__all__"))
        if assigned is not astroid.Uninferable:
            all_ = []
            for elt in getattr(assigned, "elts", ()):
                try:
                    elt_name = next(elt.infer())
                except astroid.InferenceError:
                    continue

                if elt_name is astroid.Uninferable:
                    continue

                if isinstance(elt_name, nodes.Const) and isinstance(
                    elt_name.value, str
                ):
                    all_.append(elt_name.value)

    return all_


def is_decorated_with_singledispatch(
    node: nodes.FunctionDef | nodes.AsyncFunctionDef,
) -> bool:
    """Check if the function is decorated as a singledispatch."""
    if not node.decorators:
        return False

    for decorator in node.decorators.nodes:
        if not isinstance(decorator, astroid.Name):
            continue

        try:
            if is_singledispatch_decorator(decorator):
                return True
        except astroid.InferenceError:
            pass

    return False


def is_singledispatch_decorator(decorator: astroid.Name) -> bool:
    """Check if the decorator is a singledispatch."""

    def _is_singledispatch_func(func_node: nodes.FunctionDef) -> bool:
        return (  # type: ignore[no-any-return]
            func_node.name == "singledispatch" and func_node.root().name == "functools"
        )

    for inferred in decorator.infer():
        if not isinstance(inferred, nodes.FunctionDef):
            continue

        if _is_singledispatch_func(inferred):
            return True

    return False


def is_decorated_as_singledispatch_register(
    node: nodes.FunctionDef | nodes.AsyncFunctionDef,
) -> bool:
    """Check if the function is decorated as a singledispatch register."""
    if not node.decorators:
        return False

    for decorator in node.decorators.nodes:
        if not isinstance(decorator, nodes.Call):
            continue
        if not isinstance(decorator.func, nodes.Attribute):
            continue
        if decorator.func.attrname == "register":
            return True
        # TODO any more checking?

    return False


def is_decorated_with_property(
    node: nodes.FunctionDef | nodes.AsyncFunctionDef,
) -> bool:
    """Check if the function is decorated as a property."""
    if not node.decorators:
        return False

    for decorator in node.decorators.nodes:
        if not isinstance(decorator, astroid.Name):
            continue

        try:
            if is_property_decorator(decorator):
                return True
        except astroid.InferenceError:
            pass

    return False


def is_property_decorator(decorator: astroid.Name) -> bool:
    """Check if the decorator is a property."""

    def _is_property_class(class_node: nodes.ClassDef) -> bool:
        return (  # type: ignore[no-any-return]
            class_node.name == "property"
            and class_node.root().name == builtins.__name__
        )

    for inferred in decorator.infer():
        if not isinstance(inferred, nodes.ClassDef):
            continue

        if _is_property_class(inferred):
            return True

        if any(_is_property_class(ancestor) for ancestor in inferred.ancestors()):
            return True

    return False


def is_decorated_with_property_setter(
    node: nodes.FunctionDef | nodes.AsyncFunctionDef,
) -> bool:
    """Check if the function is decorated as a property setter.

    :param node: The node to check.

    :returns: True if the function is a property setter, False otherwise.
    """
    if not node.decorators:
        return False

    for decorator in node.decorators.nodes:
        if (
            isinstance(decorator, astroid.nodes.Attribute)
            and decorator.attrname == "setter"
        ):
            return True

    return False


def get_class_docstring(node: nodes.ClassDef) -> str:
    """Get the docstring of a node, using a parent docstring if needed."""
    doc = node.doc

    if doc is None:
        for base in node.ancestors():
            if base.qname() in (
                "__builtins__.object",
                "builtins.object",
                "builtins.type",
            ):
                continue
            if base.doc is not None:
                return str(base.doc)

    return doc or ""


def is_exception(node: nodes.ClassDef) -> bool:
    """Check if a class is an exception."""
    if node.name in ("Exception", "BaseException") and node.root().name == "builtins":
        return True

    if not hasattr(node, "ancestors"):
        return False

    return any(is_exception(parent) for parent in node.ancestors(recurs=True))


def is_decorated_with_overload(node: nodes.FunctionDef) -> bool:
    """Check if the function is decorated as an overload definition."""
    if not node.decorators:
        return False

    for decorator in node.decorators.nodes:
        if not isinstance(decorator, (astroid.Name, astroid.Attribute)):
            continue

        try:
            if is_overload_decorator(decorator):
                return True
        except astroid.InferenceError:
            pass

    return False


def is_overload_decorator(decorator: astroid.Name | astroid.Attribute) -> bool:
    for inferred in decorator.infer():
        if not isinstance(inferred, astroid.nodes.FunctionDef):
            continue

        if inferred.name == "overload" and inferred.root().name == "typing":
            return True

    return False


def get_func_docstring(node: nodes.FunctionDef) -> str:
    """Get the docstring of a node, using a parent docstring if needed."""
    doc = node.doc

    if doc is None and isinstance(node.parent, nodes.ClassDef):
        for base in node.parent.ancestors():
            if node.name in ("__init__", "__new__") and base.qname() in (
                "__builtins__.object",
                "builtins.object",
                "builtins.type",
            ):
                continue
            for child in base.get_children():
                if (
                    isinstance(child, node.__class__)
                    and child.name == node.name
                    and child.doc is not None
                ):
                    return str(child.doc)

    return doc or ""


def get_return_annotation(node: nodes.FunctionDef) -> None | str:
    """Get the return annotation of a node."""
    if node.returns:
        return resolve_annotation(node.returns)
    if node.type_comment_returns:
        return resolve_annotation(node.type_comment_returns)

    return None


def get_args_info(
    args_node: astroid.Arguments,
) -> list[tuple[None | str, None | str, None | str, None | str]]:
    """Get the arguments of a function.

    :returns: a list of (type, name, annotation, default)
    """
    result: list[tuple[None | str, None | str, None | str, None | str]] = []
    positional_only_defaults = []
    positional_or_keyword_defaults = args_node.defaults
    if args_node.defaults:
        args = args_node.args or []
        positional_or_keyword_defaults = args_node.defaults[-len(args) :]
        positional_only_defaults = args_node.defaults[
            : len(args_node.defaults) - len(args)
        ]

    plain_annotations = args_node.annotations or ()

    func_comment_annotations = args_node.parent.type_comment_args or []
    if args_node.parent.type in ("method", "classmethod"):
        func_comment_annotations = [None, *func_comment_annotations]

    comment_annotations = args_node.type_comment_posonlyargs
    comment_annotations += args_node.type_comment_args or []
    comment_annotations += args_node.type_comment_kwonlyargs
    annotations = list(
        _merge_annotations(
            plain_annotations,
            _merge_annotations(func_comment_annotations, comment_annotations),
        )
    )
    annotation_offset = 0

    if args_node.posonlyargs:
        posonlyargs_annotations = args_node.posonlyargs_annotations
        if not any(args_node.posonlyargs_annotations):
            num_args = len(args_node.posonlyargs)
            posonlyargs_annotations = annotations[
                annotation_offset : annotation_offset + num_args
            ]

        for arg, annotation, default in _iter_args(
            args_node.posonlyargs, posonlyargs_annotations, positional_only_defaults
        ):
            result.append((None, arg, annotation, default))

        result.append(("/", None, None, None))

        if not any(args_node.posonlyargs_annotations):
            annotation_offset += num_args

    if args_node.args:
        num_args = len(args_node.args)
        for arg, annotation, default in _iter_args(
            args_node.args,
            annotations[annotation_offset : annotation_offset + num_args],
            positional_or_keyword_defaults,
        ):
            result.append((None, arg, annotation, default))

        annotation_offset += num_args

    if args_node.vararg:
        annotation = None
        if args_node.varargannotation:
            annotation = resolve_annotation(args_node.varargannotation)
        elif len(annotations) > annotation_offset and annotations[annotation_offset]:
            annotation = resolve_annotation(annotations[annotation_offset])
            annotation_offset += 1
        result.append(("*", args_node.vararg, annotation, None))

    if args_node.kwonlyargs:
        if not args_node.vararg:
            result.append(("*", None, None, None))

        kwonlyargs_annotations = args_node.kwonlyargs_annotations
        if not any(args_node.kwonlyargs_annotations):
            num_args = len(args_node.kwonlyargs)
            kwonlyargs_annotations = annotations[
                annotation_offset : annotation_offset + num_args
            ]

        for arg, annotation, default in _iter_args(
            args_node.kwonlyargs,
            kwonlyargs_annotations,
            args_node.kw_defaults,
        ):
            result.append((None, arg, annotation, default))

        if not any(args_node.kwonlyargs_annotations):
            annotation_offset += num_args

    if args_node.kwarg:
        annotation = None
        if args_node.kwargannotation:
            annotation = resolve_annotation(args_node.kwargannotation)
        elif len(annotations) > annotation_offset and annotations[annotation_offset]:
            annotation = resolve_annotation(annotations[annotation_offset])
            annotation_offset += 1
        result.append(("**", args_node.kwarg, annotation, None))

    if args_node.parent.type in ("method", "classmethod") and result:
        result.pop(0)

    return result


def _iter_args(
    args: list[nodes.NodeNG],
    annotations: list[nodes.NodeNG],
    defaults: list[nodes.NodeNG],
) -> t.Iterable[t.Tuple[str, None | str, str | None]]:
    """Iterate over arguments."""
    default_offset = len(args) - len(defaults)
    packed = itertools.zip_longest(args, annotations)
    for i, (arg, annotation) in enumerate(packed):
        default = None
        if (defaults is not None and i >= default_offset) and (
            defaults[i - default_offset] is not None
        ):
            default = defaults[i - default_offset].as_string()

        name = arg.name
        if isinstance(arg, astroid.Tuple):
            argument_names = ", ".join(x.name for x in arg.elts)
            name = f"({argument_names})"

        yield (
            name,
            resolve_annotation(annotation) if annotation else None,
            default,
        )


def _merge_annotations(
    annotations: t.Iterable[t.Any], comment_annotations: t.Iterable[t.Any]
) -> t.Iterable[t.Any]:
    for ann, comment_ann in itertools.zip_longest(annotations, comment_annotations):
        if ann and not _is_ellipsis(ann):
            yield ann
        elif comment_ann and not _is_ellipsis(comment_ann):
            yield comment_ann
        else:
            yield None


def _is_ellipsis(node: t.Any) -> bool:
    return isinstance(node, astroid.Const) and node.value == Ellipsis



"""Analyse a python project and create documentation for it."""

__version__ = "0.1.0"


def setup(app):  # type: ignore
    """Entrypoint for sphinx."""
    from .extension import setup as _setup

    return _setup(app)



"""Analyse of Python code using astroid.

The core function though `analyse_module` is agnostic to the implementation,
It simply yields `ItemData` typed-dicts.
"""
from __future__ import annotations

from dataclasses import dataclass, replace
import itertools
import os
from pathlib import Path
import sys
import typing as t

from astroid import nodes
from astroid.builder import AstroidBuilder

from . import astroid_utils

if t.TYPE_CHECKING:
    from .utils import PROPERTY_TYPE, ItemData

__all__ = ["analyse_module"]


def analyse_module(
    file_path: Path, name: str, exclude_external_imports: t.Pattern[str] | None = None
) -> t.Iterable[ItemData]:
    """Analyse the given module and yield items.

    :param file_path: The path to the module.
    :param name: The name of the module, e.g. "foo.bar".
    :param record_external_imports: If given, record these external imports on the module.
        These are only used to determine what is exposed by __all__,
        which is only usually objects in the same package.
        But if you want to expose objects from other packages,
        you can use this to record them.
    """
    # TODO expose record_external_imports everywhere analyse_module is used
    node = AstroidBuilder().file_build(os.fsdecode(file_path), name)
    yield from walk_node(
        node, State(node.name.split(".", 1)[0], [], exclude_external_imports)
    )


_dc_kwargs: dict[str, bool] = {"frozen": True}
if sys.version_info >= (3, 10):
    _dc_kwargs["slots"] = True


@dataclass(**_dc_kwargs)
class State:
    package_name: str
    name_stack: list[str]
    exclude_external_imports: t.Pattern[str] | None

    def copy(self, **kwargs: t.Any) -> State:
        """Copy the state and update the given attributes."""
        return replace(self, **kwargs)


def _get_full_name(name: str, name_stack: list[str]) -> str:
    """Get the full name of a node."""
    return ".".join([*name_stack, name])


def _get_parent_name(name: str) -> str:
    """Get the parent name of a node."""
    return ".".join(name.split(".")[:-1])


def clean_docstring(s: None | str, tabsize: int = 8) -> str:
    """Remove common leading indentation,
    where the indentation of the first line is ignored.
    """
    if s is None:
        return ""
    lines = s.expandtabs(tabsize).splitlines()
    # Find minimum indentation of any non-blank lines after ignored lines.
    margin = sys.maxsize
    for line in lines[1:]:
        content = len(line.lstrip())
        if content:
            indent = len(line) - content
            margin = min(margin, indent)
    # Remove indentation from the first line.
    if len(lines):
        lines[0] = lines[0].lstrip()
    if margin < sys.maxsize:
        for i in range(1, len(lines)):
            lines[i] = lines[i][margin:]
    # Remove any leading blank lines.
    while lines and not lines[0]:
        lines.pop(0)

    return "\n".join(lines)


def walk_node(node: nodes.NodeNG, state: State) -> t.Iterable[ItemData]:
    func = _FUNC_MAPPER.get(type(node))
    if func is not None:
        yield from func(node, state)


def yield_module(node: nodes.Module, state: State) -> t.Iterable[ItemData]:
    path = node.path
    if isinstance(node.path, list):
        path = node.path[0] if node.path else None

    parent: ItemData = {
        "type": "package" if node.package else "module",
        "full_name": node.name,
        "doc": clean_docstring(node.doc),
        "file_path": path,
        "encoding": node.file_encoding,
        "all": astroid_utils.get_module_all(node),
    }

    for child in node.get_children():
        if isinstance(child, nodes.ImportFrom):
            # Note, the code below restricts to local imports, which was in sphinx-autoapi
            # However, we potentially need all imports, to be able to resolve __all__ exports
            # (i.e. the public API) since packages could expose aspects of other packages
            # To limit what we store in the database, we allow for filtering out imports
            # for example, it might be reasonably assumed that nothing imported from the typing module
            # is part of the public API
            if not (
                child.level
                or child.modname == state.package_name
                or child.modname.startswith(state.package_name + ".")
            ) and (
                child.modname
                and state.exclude_external_imports
                and state.exclude_external_imports.fullmatch(child.modname)
            ):
                continue

            imports = parent.setdefault("imports", [])
            for name, alias in child.names:
                original_path = astroid_utils.get_full_import_name(child, alias or name)
                imports.append((original_path, alias))

    yield parent

    for child in node.get_children():
        yield from walk_node(child, state.copy(name_stack=[node.name]))


def yield_annotation_assign(
    node: nodes.AnnAssign, state: State
) -> t.Iterable[ItemData]:
    """Yield data for an annotation assignment node."""
    if not isinstance(node.target, nodes.AssignAttr):
        yield from _yield_assign(node, state)


def yield_assign(node: nodes.Assign, state: State) -> t.Iterable[ItemData]:
    """Yield data for an assignment node."""
    if not any(isinstance(target, nodes.AssignAttr) for target in node.targets):
        yield from _yield_assign(node, state)


def _yield_assign(
    node: nodes.Assign | nodes.AnnAssign, state: State
) -> t.Iterable[ItemData]:
    """Yield data for an assignment node."""
    doc = ""
    doc_node = node.next_sibling()
    if isinstance(doc_node, nodes.Expr) and isinstance(doc_node.value, nodes.Const):
        doc = doc_node.value.value

    type_ = "data"
    if isinstance(node.scope(), nodes.ClassDef) or astroid_utils.is_constructor(
        node.scope()
    ):
        type_ = "attribute"

    assign_value = astroid_utils.get_assign_value(node)
    if not assign_value:
        return

    target = assign_value[0]
    value = assign_value[1]

    annotation = astroid_utils.get_assign_annotation(node)

    data: ItemData = {
        "type": type_,
        "full_name": _get_full_name(target, state.name_stack),
        "doc": clean_docstring(doc),
        "value": value,
        "annotation": annotation,
    }
    if node.fromlineno is not None and node.tolineno is not None:
        data["range"] = (node.fromlineno, node.tolineno)
    yield data


def yield_function_def(
    node: nodes.FunctionDef | nodes.AsyncFunctionDef, state: State
) -> t.Iterable[ItemData]:
    """Yield data for a function definition node."""
    if astroid_utils.is_decorated_with_property_setter(node):
        return
    if astroid_utils.is_decorated_as_singledispatch_register(node):
        # functools.singledispatch registers, we just ignore these for now,
        # since they are not really functions
        return

    type_ = "method"
    properties: list[PROPERTY_TYPE] = []

    if node.type == "function":
        type_ = "function"

        if isinstance(node, nodes.AsyncFunctionDef):
            properties.append("async")
        if astroid_utils.is_decorated_with_singledispatch(node):
            properties.append("singledispatch")
    elif astroid_utils.is_decorated_with_property(node):
        type_ = "property"
        if node.type == "classmethod":
            properties.append(node.type)
        if node.is_abstract(pass_is_abstract=False):
            properties.append("abstractmethod")
    else:
        # "__new__" method is implicit classmethod
        if node.type in ("staticmethod", "classmethod") and node.name != "__new__":
            properties.append(node.type)
        if node.is_abstract(pass_is_abstract=False):
            properties.append("abstractmethod")
        if isinstance(node, nodes.AsyncFunctionDef):
            properties.append("async")

    if astroid_utils.is_decorated_with_overload(node):
        type_ = "overload"

    data: ItemData = {
        "type": type_,
        "full_name": _get_full_name(node.name, state.name_stack),
        "doc": clean_docstring(astroid_utils.get_func_docstring(node)),
        "args": astroid_utils.get_args_info(node.args),
        "return_annotation": astroid_utils.get_return_annotation(node),
    }
    if node.fromlineno is not None and node.tolineno is not None:
        data["range"] = (node.fromlineno, node.tolineno)
    if properties:
        data["properties"] = properties

    yield data

    if node.name == "__init__":
        for child in node.get_children():
            if isinstance(child, (nodes.Assign, nodes.AnnAssign)):
                child_data = _yield_assign(child, state)
                for data in child_data:
                    if data["doc"]:
                        yield data


def yield_class_def(node: nodes.ClassDef, state: State) -> t.Iterable[ItemData]:
    """Yield data for a class definition node."""
    type_ = "class"
    if astroid_utils.is_exception(node):
        type_ = "exception"

    basenames = [astroid_utils.resolve_annotation(base) for base in node.bases]

    parent: ItemData = {
        "type": type_,
        "full_name": _get_full_name(node.name, state.name_stack),
        "bases": basenames,
        "doc": clean_docstring(astroid_utils.get_class_docstring(node)),
    }
    if node.fromlineno is not None and node.tolineno is not None:
        parent["range"] = (node.fromlineno, node.tolineno)
    yield parent

    new_state = state.copy(name_stack=[*state.name_stack, node.name])

    overridden: t.Set[str] = set()  # a list of methods overridden by class inheritance
    for base in itertools.chain(iter((node,)), node.ancestors()):
        seen: t.Set[str] = set()
        if base.qname() in (
            "__builtins__.object",
            "builtins.object",
            "builtins.type",
        ):
            continue
        for child in base.get_children():
            name = getattr(child, "name", None)
            if isinstance(child, (nodes.Assign, nodes.AnnAssign)):
                assign_value = astroid_utils.get_assign_value(child)
                if not assign_value:
                    continue
                name = assign_value[0]

            if not name or name in overridden:
                continue
            seen.add(name)

            for ancestor in walk_node(child, new_state):
                if (
                    _get_parent_name(ancestor["full_name"]) == parent["full_name"]
                ) and (base is not node):
                    ancestor["inherited"] = True

                yield ancestor

        overridden.update(seen)


_FUNC_MAPPER: dict[
    nodes.NodeNG, t.Callable[[nodes.NodeNG, State], t.Iterable[ItemData]]
] = {
    nodes.Module: yield_module,
    nodes.AnnAssign: yield_annotation_assign,
    nodes.Assign: yield_assign,
    nodes.FunctionDef: yield_function_def,
    nodes.AsyncFunctionDef: yield_function_def,
    nodes.ClassDef: yield_class_def,
}



"""Renderer for MyST."""
from __future__ import annotations

import re
import typing as t

from autodoc2.render.base import RendererBase, format_args

if t.TYPE_CHECKING:
    from autodoc2.utils import ItemData


class MystRenderer(RendererBase):
    """Render the documentation as MyST."""

    EXTENSION = ".md"

    def render_item(self, full_name: str) -> t.Iterable[str]:
        item = self.get_item(full_name)
        if item is None:
            raise ValueError(f"Item {full_name} does not exist")
        type_ = item["type"]
        if type_ == "package":
            yield from self.render_package(item)
        elif type_ == "module":
            yield from self.render_module(item)
        elif type_ == "function":
            yield from self.render_function(item)
        elif type_ == "class":
            yield from self.render_class(item)
        elif type_ == "exception":
            yield from self.render_exception(item)
        elif type_ == "property":
            yield from self.render_property(item)
        elif type_ == "method":
            yield from self.render_method(item)
        elif type_ == "attribute":
            yield from self.render_attribute(item)
        elif type_ == "data":
            yield from self.render_data(item)
        else:
            self.warn(f"Unknown item type {type_!r} for {full_name!r}")

    def generate_summary(self, items: list[ItemData]) -> t.Iterable[str]:
        """Generate a summary of the items."""
        for item in items:
            short_name = item["full_name"].split(".")[-1]
            summary = ""
            # get the first paragraph, allowing for the first line to be blank
            for i, line in enumerate(item["doc"].splitlines()):
                if i > 0 and not line.strip():
                    break
                summary += line.rstrip() + " "
            summary = summary.strip()
            # TODO get signature (for functions, etc), plus sphinx also runs rst.escape
            yield f"* - {{py:obj}}`{short_name} <{item['full_name']}>`"
            yield f"  - {summary}"

    @staticmethod
    def enclosing_backticks(text: str) -> str:
        """Ensure the enclosing backticks are more than any inner ones."""
        backticks = "```"
        while backticks in text:
            backticks += "`"
        return backticks

    def render_package(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a package."""
        if self.is_hidden(item):
            yield from ["---", "orphan: true", "---", ""]

        full_name = item["full_name"]

        yield f"# {{py:mod}}`{full_name}`"
        yield ""

        yield f"```{{py:module}} {full_name}"
        if self.is_module_deprecated(item):
            yield ":deprecated:"
        yield from ["```", ""]

        if item["doc"]:
            yield "## Description"
            yield ""
            for line in item["doc"].splitlines():
                if _MD_HEAD_RE.match(line):
                    # if its a heading, then we want to make sure it will be
                    # a sub heading of description
                    yield "##" + line
                else:
                    yield line
            yield ""

        visible_subpackages = [
            i["full_name"] for i in self.get_children(item, {"package"})
        ]
        if visible_subpackages:
            yield from [
                "## Subpackages",
                "",
                "```{toctree}",
                ":titlesonly:",
                ":maxdepth: 3",
                "",
            ]
            for name in visible_subpackages:
                yield name
            yield "```"
            yield ""

        visible_submodules = [
            i["full_name"] for i in self.get_children(item, {"module"})
        ]
        if visible_submodules:
            yield from [
                "## Submodules",
                "",
                "```{toctree}",
                ":titlesonly:",
                ":maxdepth: 1",
                "",
            ]
            for name in visible_submodules:
                yield name
            yield "```"
            yield ""

        visible_children = [
            i["full_name"]
            for i in self.get_children(item)
            if i["type"] not in ("package", "module")
        ]
        if not visible_children:
            return

        yield f"## {item['type'].capitalize()} Contents"
        yield ""

        if self.show_module_summary(item):
            for heading, types in [
                ("Classes", {"class"}),
                ("Functions", {"function"}),
                ("Data", {"data"}),
                ("External", {"external"}),
            ]:
                visible_items = list(self.get_children(item, types))
                if visible_items:
                    yield from [
                        f"### {heading}",
                        "",
                        "```{list-table}",
                        ":class: autosummary longtable",
                        ":align: left",
                        "",
                    ]
                    yield from self.generate_summary(visible_items)
                    yield "```"
                    yield ""

            yield from ["### API", ""]
            for name in visible_children:
                yield from self.render_item(name)

    def render_module(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a module."""
        yield from self.render_package(item)

    def render_function(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a function."""
        if self.is_hidden(item):
            return

        short_name = item["full_name"].split(".")[-1]
        show_annotations = self.show_annotations(item)
        sig = f"{short_name}({format_args(item['args'], show_annotations)})"
        if show_annotations and item.get("return_annotation"):
            sig += f" -> {item['return_annotation']}"

        backticks = self.enclosing_backticks(item["doc"] or "")

        yield f"{backticks}{{py:function}} {sig}"
        yield f":canonical: {item['full_name']}"
        # TODO overloads
        if "async" in item.get("properties", []):
            yield ":async:"
            # TODO it would also be good to highlight if singledispatch decorated,
            # or, more broadly speaking, decorated at all
        yield ""
        if item["doc"]:
            yield item["doc"]
            yield ""
        yield backticks
        yield ""

    def render_exception(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for an exception."""
        yield from self.render_class(item)

    def render_class(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a class."""
        if self.is_hidden(item):
            return

        short_name = item["full_name"].split(".")[-1]
        constructor = self.get_item(f"{item['full_name']}.__init__")
        sig = short_name
        if constructor and "args" in constructor:
            sig += f"({format_args(constructor['args'], self.show_annotations(item), ignore_self='self')})"

        # note, here we can cannot yield by line,
        # because we need to look ahead to know the length of the backticks

        lines: list[str] = []

        lines += [
            f":canonical: {item['full_name']}",
            "",
        ]

        # TODO overloads

        if item.get("bases") and self.show_class_inheritance(item):
            lines += [
                "Bases: "
                + ", ".join(
                    [_reformat_cls_base_myst(b) for b in item.get("bases", [])]
                ),
                "",
            ]

        # TODO inheritance diagram

        if item["doc"]:
            lines += [item["doc"], ""]

        for child in self.get_children(
            item, {"class", "property", "attribute", "method"}
        ):
            for line in self.render_item(child["full_name"]):
                lines.append(line)

        backticks = self.enclosing_backticks("\n".join(lines))
        yield f"{backticks}{{py:{item['type']}}} {sig}"
        for line in lines:
            yield line
        yield backticks
        yield ""

    def render_property(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a property."""
        if self.is_hidden(item):
            return

        backticks = self.enclosing_backticks(item["doc"] or "")

        short_name = item["full_name"].split(".")[-1]
        yield f"{backticks}{{py:property}} {short_name}"
        yield f":canonical: {item['full_name']}"
        for prop in ("abstractmethod", "classmethod"):
            if prop in item.get("properties", []):
                yield f":{prop}:"
        if item.get("return_annotation"):
            yield f":type: {item['return_annotation']}"
        yield ""

        if item["doc"]:
            yield item["doc"]
            yield ""

        yield backticks
        yield ""

    def render_method(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a method."""
        if self.is_hidden(item):
            return

        short_name = item["full_name"].split(".")[-1]
        show_annotations = self.show_annotations(item)
        sig = f"{short_name}({format_args(item['args'], show_annotations, ignore_self='self')})"
        if show_annotations and item.get("return_annotation"):
            sig += f" -> {item['return_annotation']}"

        backticks = self.enclosing_backticks(item["doc"] or "")

        yield f"{backticks}{{py:method}} {sig}"
        yield f":canonical: {item['full_name']}"
        # TODO overloads
        # TODO collect final decorated in analysis
        for prop in ("abstractmethod", "async", "classmethod", "final", "staticmethod"):
            if prop in item.get("properties", []):
                yield f":{prop}:"
        yield ""

        if item["doc"]:
            yield item["doc"]
            yield ""

        yield backticks
        yield ""

    def render_attribute(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for an attribute."""
        yield from self.render_data(item)

    def render_data(self, item: ItemData) -> t.Iterable[str]:
        """Create the content for a data item."""
        if self.is_hidden(item):
            return

        backticks = self.enclosing_backticks(item["doc"] or "")
        short_name = item["full_name"].split(".")[-1]

        yield f"{backticks}{{py:{item['type']}}} {short_name}"
        yield f":canonical: {item['full_name']}"
        for prop in ("abstractmethod", "classmethod"):
            if prop in item.get("properties", []):
                yield f":{prop}:"
        if item.get("annotation"):
            yield f":type: {item['annotation']}"
        value = item.get("value")
        if isinstance(value, str):
            if len(value.splitlines()) == 1:
                if len(value) > 100:
                    value = value[:100] + "..."
                yield ":value: >"  # use > to ensure its understood as a string
                yield f"   {value!r}"
            else:
                yield ":value: <Multiline-String>"
                # TODO in sphinx-autoapi, they made a code block inside a details/summary HTML
        else:
            value = str(value).replace("\n", " ")
            if len(value) > 100:
                value = value[:100] + "..."
            yield ":value: >"
            yield f"   {value}"

        yield ""
        if item["doc"]:
            yield item["doc"]
            yield ""

        yield backticks
        yield ""


_MD_HEAD_RE = re.compile(r"^[#]+\s")
_RE_DELIMS = re.compile(r"(\s*[\[\]\(\),]\s*)")


def _reformat_cls_base_myst(value: str) -> str:
    """Reformat the base of a class for RST.

    Base annotations can come in the form::

        A[B, C, D]

    which we want to reformat as::

        {py:obj}`A`\\[{py:obj}`B`, {py:obj}`C`, {py:obj}`D`\\]

    """
    result = ""
    for sub_target in _RE_DELIMS.split(value.strip()):
        sub_target = sub_target.strip()
        if _RE_DELIMS.match(sub_target):
            result += f"{sub_target}"
            if sub_target.endswith(","):
                result += " "
            else:
                result += "\\"
        elif sub_target:
            if result.endswith("\\"):
                result = result[:-1]
            result += f"{{py:obj}}`{sub_target}`\\"

    if result.endswith("\\"):
        result = result[:-1]

    return result
